'''
This module defines :class:`Segment`, a container for data sharing a common
time basis.

:class:`Segment` derives from :class:`Container`,
from :module:`neo.core.container`.
'''

from datetime import datetime

import numpy as np

from copy import deepcopy

from neo.core.container import Container
from neo.core.spiketrainlist import SpikeTrainList


class Segment(Container):
    '''
    A container for data sharing a common time basis.

    A :class:`Segment` is a heterogeneous container for discrete or continous
    data sharing a common clock (time basis) but not necessary the same
    sampling rate, start or end time.

    *Usage*::
        >>> from neo.core import Segment, SpikeTrain, AnalogSignal
        >>> from quantities import Hz, s
        >>>
        >>> seg = Segment(index=5)
        >>>
        >>> train0 = SpikeTrain(times=[.01, 3.3, 9.3], units='sec', t_stop=10)
        >>> seg.spiketrains.append(train0)
        >>>
        >>> train1 = SpikeTrain(times=[100.01, 103.3, 109.3], units='sec',
        ...                     t_stop=110)
        >>> seg.spiketrains.append(train1)
        >>>
        >>> sig0 = AnalogSignal(signal=[.01, 3.3, 9.3], units='uV',
        ...                     sampling_rate=1*Hz)
        >>> seg.analogsignals.append(sig0)
        >>>
        >>> sig1 = AnalogSignal(signal=[100.01, 103.3, 109.3], units='nA',
        ...                     sampling_period=.1*s)
        >>> seg.analogsignals.append(sig1)

    *Required attributes/properties*:
        None

    *Recommended attributes/properties*:
        :name: (str) A label for the dataset.
        :description: (str) Text description.
        :file_origin: (str) Filesystem path or URL of the original data file.
        :file_datetime: (datetime) The creation date and time of the original
            data file.
        :rec_datetime: (datetime) The date and time of the original recording
        :index: (int) You can use this to define a temporal ordering of
            your Segment. For instance you could use this for trial numbers.

    Note: Any other additional arguments are assumed to be user-specific
    metadata and stored in :attr:`annotations`.

    *Properties available on this object*:
        :all_data: (list) A list of all child objects in the :class:`Segment`.

    *Container of*:
        :class:`Epoch`
        :class:`Event`
        :class:`AnalogSignal`
        :class:`IrregularlySampledSignal`
        :class:`SpikeTrain`

    '''

    _data_child_objects = ('AnalogSignal',
                           'Epoch', 'Event',
                           'IrregularlySampledSignal', 'SpikeTrain', 'ImageSequence')
    _parent_objects = ('Block',)
    _recommended_attrs = ((('file_datetime', datetime),
                           ('rec_datetime', datetime),
                           ('index', int)) +
                          Container._recommended_attrs)
    _repr_pretty_containers = ('analogsignals',)

    def __init__(self, name=None, description=None, file_origin=None,
                 file_datetime=None, rec_datetime=None, index=None,
                 **annotations):
        '''
        Initialize a new :class:`Segment` instance.
        '''
        super().__init__(name=name, description=description,
                         file_origin=file_origin, **annotations)
        self.spiketrains = SpikeTrainList(segment=self)
        self.file_datetime = file_datetime
        self.rec_datetime = rec_datetime
        self.index = index

    # t_start attribute is handled as a property so type checking can be done
    @property
    def t_start(self):
        '''
        Time when first signal begins.
        '''
        t_starts = [sig.t_start for sig in self.analogsignals +
                    self.spiketrains + self.irregularlysampledsignals]

        for e in self.epochs + self.events:
            if hasattr(e, 't_start'):  # in case of proxy objects
                t_starts += [e.t_start]
            elif len(e) > 0:
                t_starts += [e.times[0]]

        # t_start is not defined if no children are present
        if len(t_starts) == 0:
            return None

        t_start = min(t_starts)
        return t_start

    # t_stop attribute is handled as a property so type checking can be done
    @property
    def t_stop(self):
        '''
        Time when last signal ends.
        '''
        t_stops = [sig.t_stop for sig in self.analogsignals +
                   self.spiketrains + self.irregularlysampledsignals]

        for e in self.epochs + self.events:
            if hasattr(e, 't_stop'):  # in case of proxy objects
                t_stops += [e.t_stop]
            elif len(e) > 0:
                t_stops += [e.times[-1]]

        # t_stop is not defined if no children are present
        if len(t_stops) == 0:
            return None

        t_stop = max(t_stops)
        return t_stop

    def time_slice(self, t_start=None, t_stop=None, reset_time=False, **kwargs):
        """
        Creates a time slice of a Segment containing slices of all child
        objects.

        Parameters:
        -----------
        t_start: Quantity
            Starting time of the sliced time window.
        t_stop: Quantity
            Stop time of the sliced time window.
        reset_time: bool
            If True the time stamps of all sliced objects are set to fall
            in the range from t_start to t_stop.
            If False, original time stamps are retained.
            Default is False.

        Keyword Arguments:
        ------------------
            Additional keyword arguments used for initialization of the sliced
            Segment object.

        Returns:
        --------
        subseg: Segment
            Temporal slice of the original Segment from t_start to t_stop.
        """
        subseg = Segment(**kwargs)

        for attr in ['file_datetime', 'rec_datetime', 'index',
                     'name', 'description', 'file_origin']:
            setattr(subseg, attr, getattr(self, attr))

        subseg.annotations = deepcopy(self.annotations)

        if t_start is None:
            t_start = self.t_start
        if t_stop is None:
            t_stop = self.t_stop

        t_shift = - t_start

        # cut analogsignals and analogsignalarrays
        for ana_id in range(len(self.analogsignals)):
            if hasattr(self.analogsignals[ana_id], '_rawio'):
                ana_time_slice = self.analogsignals[ana_id].load(time_slice=(t_start, t_stop))
            else:
                ana_time_slice = self.analogsignals[ana_id].time_slice(t_start, t_stop)
            if reset_time:
                ana_time_slice = ana_time_slice.time_shift(t_shift)
            subseg.analogsignals.append(ana_time_slice)

        # cut irregularly sampled signals
        for irr_id in range(len(self.irregularlysampledsignals)):
            if hasattr(self.irregularlysampledsignals[irr_id], '_rawio'):
                ana_time_slice = self.irregularlysampledsignals[irr_id].load(
                    time_slice=(t_start, t_stop))
            else:
                ana_time_slice = self.irregularlysampledsignals[irr_id].time_slice(t_start, t_stop)
            if reset_time:
                ana_time_slice = ana_time_slice.time_shift(t_shift)
            subseg.irregularlysampledsignals.append(ana_time_slice)

        # cut spiketrains
        for st_id in range(len(self.spiketrains)):
            if hasattr(self.spiketrains[st_id], '_rawio'):
                st_time_slice = self.spiketrains[st_id].load(time_slice=(t_start, t_stop))
            else:
                st_time_slice = self.spiketrains[st_id].time_slice(t_start, t_stop)
            if reset_time:
                st_time_slice = st_time_slice.time_shift(t_shift)
            subseg.spiketrains.append(st_time_slice)

        # cut events
        for ev_id in range(len(self.events)):
            if hasattr(self.events[ev_id], '_rawio'):
                ev_time_slice = self.events[ev_id].load(time_slice=(t_start, t_stop))
            else:
                ev_time_slice = self.events[ev_id].time_slice(t_start, t_stop)
            if reset_time:
                ev_time_slice = ev_time_slice.time_shift(t_shift)
            # appending only non-empty events
            if len(ev_time_slice):
                subseg.events.append(ev_time_slice)

        # cut epochs
        for ep_id in range(len(self.epochs)):
            if hasattr(self.epochs[ep_id], '_rawio'):
                ep_time_slice = self.epochs[ep_id].load(time_slice=(t_start, t_stop))
            else:
                ep_time_slice = self.epochs[ep_id].time_slice(t_start, t_stop)
            if reset_time:
                ep_time_slice = ep_time_slice.time_shift(t_shift)
            # appending only non-empty epochs
            if len(ep_time_slice):
                subseg.epochs.append(ep_time_slice)

        subseg.create_relationship()

        return subseg



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.phyrawio import PhyRawIO


class PhyIO(PhyRawIO, BaseFromRaw):
    name = 'Phy IO'
    description = "Phy IO"
    mode = 'dir'

    def __init__(self, dirname):
        PhyRawIO.__init__(self, dirname=dirname)
        BaseFromRaw.__init__(self, dirname)



"""
Tests of neo.io.spikeglxio
"""

import unittest

from neo.io import SpikeGLXIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestSpikeGLXIO(BaseTestIO, unittest.TestCase):
    ioclass = SpikeGLXIO
    entities_to_download = [
        'spikeglx'
    ]
    entities_to_test = [
        'spikeglx/Noise4Sam_g0',
        'spikeglx/TEST_20210920_0_g0'
    ]
    


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.alphaomegaio
"""

import unittest

from neo.io import AlphaOmegaIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestAlphaOmegaIO(BaseTestIO, unittest.TestCase):
    entities_to_download =[
        'alphaomega'
    ]
    entities_to_test = [
        'alphaomega/File_AlphaOmega_1.map',
        'alphaomega/File_AlphaOmega_2.map'
    ]
    ioclass = AlphaOmegaIO


if __name__ == "__main__":
    unittest.main()



'''
This module defines :class:`Block`, the main container gathering all the data,
whether discrete or continous, for a given recording session. base class
used by all :module:`neo.core` classes.

:class:`Block` derives from :class:`Container`,
from :module:`neo.core.container`.
'''

from datetime import datetime

from neo.core.container import Container, unique_objs


class Block(Container):
    '''
    Main container gathering all the data, whether discrete or continous, for a
    given recording session.

    A block is not necessarily temporally homogeneous, in contrast to :class:`Segment`.

    *Usage*::

        >>> from neo.core import Block, Segment, Group, AnalogSignal
        >>> from quantities import nA, kHz
        >>> import numpy as np
        >>>
        >>> # create a Block with 3 Segment and 2 Group objects
        ,,, blk = Block()
        >>> for ind in range(3):
        ...     seg = Segment(name='segment %d' % ind, index=ind)
        ...     blk.segments.append(seg)
        ...
        >>> for ind in range(2):
        ...     group = Group(name='Array probe %d' % ind)
        ...     blk.groups.append(group)
        ...
        >>> # Populate the Block with AnalogSignal objects
        ... for seg in blk.segments:
        ...     for group in blk.groups:
        ...         a = AnalogSignal(np.random.randn(10000, 64)*nA,
        ...                          sampling_rate=10*kHz)
        ...         group.analogsignals.append(a)
        ...         seg.analogsignals.append(a)

    *Required attributes/properties*:
        None

    *Recommended attributes/properties*:
        :name: (str) A label for the dataset.
        :description: (str) Text description.
        :file_origin: (str) Filesystem path or URL of the original data file.
        :file_datetime: (datetime) The creation date and time of the original
            data file.
        :rec_datetime: (datetime) The date and time of the original recording.

    Note: Any other additional arguments are assumed to be user-specific
    metadata and stored in :attr:`annotations`.

    *Container of*:
        :class:`Segment`
        :class:`Group`

    '''

    _container_child_objects = ('Segment', 'Group')
    _child_properties = ()
    _recommended_attrs = ((('file_datetime', datetime),
                           ('rec_datetime', datetime),
                           ('index', int)) +
                          Container._recommended_attrs)
    _repr_pretty_attrs_keys_ = (Container._repr_pretty_attrs_keys_ +
                                ('file_origin', 'file_datetime',
                                 'rec_datetime', 'index'))
    _repr_pretty_containers = ('segments',)

    def __init__(self, name=None, description=None, file_origin=None,
                 file_datetime=None, rec_datetime=None, index=None,
                 **annotations):
        '''
        Initalize a new :class:`Block` instance.
        '''
        super().__init__(name=name, description=description,
                                    file_origin=file_origin, **annotations)

        self.file_datetime = file_datetime
        self.rec_datetime = rec_datetime
        self.index = index
        self.regionsofinterest = []   # temporary workaround.
        # the goal is to store all sub-classes of RegionOfInterest in a single list
        # but this will need substantial changes to container handling

    @property
    def data_children_recur(self):
        '''
        All data child objects stored in the current object,
        obtained recursively.
        '''
        # subclassing this to remove duplicate objects such as SpikeTrain
        # objects in both Segment and Group
        # Only Block can have duplicate items right now, so implement
        # this here for performance reasons.
        return tuple(unique_objs(super().data_children_recur))

    def list_children_by_class(self, cls):
        '''
        List all children of a particular class recursively.

        You can either provide a class object, a class name,
        or the name of the container storing the class.
        '''
        # subclassing this to remove duplicate objects such as SpikeTrain
        # objects in both Segment and Group
        # Only Block can have duplicate items right now, so implement
        # this here for performance reasons.
        return unique_objs(super().list_children_by_class(cls))



"""
Tests of neo.rawio.phyrawio

Author: Regimantas Jurkus

"""

import unittest

from neo.rawio.phyrawio import PhyRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

import csv
import tempfile
from pathlib import Path
from collections import OrderedDict
import sys


class TestPhyRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = PhyRawIO
    entities_to_download = [
        'phy'
    ]
    entities_to_test = [
        'phy/phy_example_0'
    ]

    def test_csv_tsv_parser_with_csv(self):
        csv_tempfile = Path(tempfile.gettempdir()).joinpath('test.csv')
        with open(csv_tempfile, 'w') as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=',')
            csv_writer.writerow(['cluster_id', 'some_annotation'])
            csv_writer.writerow([1, 'Good'])
            csv_writer.writerow([2, 10])
            csv_writer.writerow([3, 1.23])

        # the parser in PhyRawIO runs csv.DictReader to parse the file
        # csv.DictReader for python version 3.6+ returns list of OrderedDict
        if (3, 6) <= sys.version_info < (3, 8):
            target = [OrderedDict({'cluster_id': 1,
                                   'some_annotation': 'Good'}),
                      OrderedDict({'cluster_id': 2,
                                   'some_annotation': 10}),
                      OrderedDict({'cluster_id': 3,
                                   'some_annotation': 1.23})]

        # csv.DictReader for python version 3.8+ returns list of dict
        elif sys.version_info >= (3, 8):
            target = [{'cluster_id': 1, 'some_annotation': 'Good'},
                      {'cluster_id': 2, 'some_annotation': 10},
                      {'cluster_id': 3, 'some_annotation': 1.23}]

        list_of_dict = PhyRawIO._parse_tsv_or_csv_to_list_of_dict(csv_tempfile)

        self.assertEqual(target, list_of_dict)

    def test_csv_tsv_parser_with_tsv(self):
        tsv_tempfile = Path(tempfile.gettempdir()).joinpath('test.tsv')
        with open(tsv_tempfile, 'w') as tsv_file:
            tsv_writer = csv.writer(tsv_file, delimiter='\t')
            tsv_writer.writerow(['cluster_id', 'some_annotation'])
            tsv_writer.writerow([1, 'Good'])
            tsv_writer.writerow([2, 10])
            tsv_writer.writerow([3, 1.23])

        # the parser in PhyRawIO runs csv.DictReader to parse the file
        # csv.DictReader for python version 3.6+ returns list of OrderedDict
        if (3, 6) <= sys.version_info < (3, 8):
            target = [OrderedDict({'cluster_id': 1,
                                   'some_annotation': 'Good'}),
                      OrderedDict({'cluster_id': 2,
                                   'some_annotation': 10}),
                      OrderedDict({'cluster_id': 3,
                                   'some_annotation': 1.23})]

        # csv.DictReader for python version 3.8+ returns list of dict
        elif sys.version_info >= (3, 8):
            target = [{'cluster_id': 1, 'some_annotation': 'Good'},
                      {'cluster_id': 2, 'some_annotation': 10},
                      {'cluster_id': 3, 'some_annotation': 1.23}]

        list_of_dict = PhyRawIO._parse_tsv_or_csv_to_list_of_dict(tsv_tempfile)

        self.assertEqual(target, list_of_dict)

    def test_csv_tsv_parser_error_raising(self):
        txt_tempfile = Path(tempfile.gettempdir()).joinpath('test.txt')
        with open(txt_tempfile, 'w') as txt_file:
            txt_file.write('This is a test')

        self.assertRaises(ValueError,
                          PhyRawIO._parse_tsv_or_csv_to_list_of_dict,
                          txt_tempfile)


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.brainvisionio
"""

import unittest

from neo.io import BrainVisionIO

from neo.test.iotest.common_io_test import BaseTestIO


class TestBrainVisionIO(BaseTestIO, unittest.TestCase, ):
    ioclass = BrainVisionIO
    entities_to_download = [
        'brainvision'
    ]
    entities_to_test = [
        'brainvision/File_brainvision_1.vhdr',
        'brainvision/File_brainvision_2.vhdr',
        'brainvision/File_brainvision_3_float32.vhdr',
        'brainvision/File_brainvision_3_int16.vhdr',
        'brainvision/File_brainvision_3_int32.vhdr',
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.openephysrawio import OpenEphysRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestOpenEphysRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = OpenEphysRawIO
    entities_to_download = [
        'openephys'
    ]
    entities_to_test = [
        'openephys/OpenEphys_SampleData_1',
        # 'OpenEphys_SampleData_2_(multiple_starts)',  # This not implemented this raise error
        # 'OpenEphys_SampleData_3',
    ]

    def test_raise_error_if_discontinuous_files(self):
        # the case of discontinuous signals is NOT cover by the IO for the moment
        # It must raise an error
        reader = OpenEphysRawIO(dirname=self.get_local_path(
            'openephys/OpenEphys_SampleData_2_(multiple_starts)'))
        with self.assertRaises(Exception):
            reader.parse_header()

    def test_raise_error_if_strange_timestamps(self):
        # In this dataset CH32 have strange timestamps
        reader = OpenEphysRawIO(dirname=self.get_local_path('openephys/OpenEphys_SampleData_3'))
        with self.assertRaises(Exception):
            reader.parse_header()

    def test_channel_order(self):
        reader = OpenEphysRawIO(dirname=self.get_local_path(
            'openephys/OpenEphys_SampleData_1'))
        reader.parse_header()
        reader.header['signal_channels']['name'][0].startswith('CH')


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
This generate diagram of the (raw)ios and formats


Author: Julia Sprenger
"""

import pygraphviz
import neo

# from datetime import datetime
#
# import numpy as np
# import quantities as pq
# from matplotlib import pyplot
# from matplotlib.patches import Rectangle, ArrowStyle, FancyArrowPatch
# from matplotlib.font_manager import FontProperties
#
# from neo.test.generate_datasets import fake_neo
#
# line_heigth = .22
# fontsize = 10.5
# left_text_shift = .1
# dpi = 100


default_style = {'shape': 'rectangle',
                 'color': 'black',
                 'fontcolor': 'black'}
IO_style = default_style.copy()
IO_style['fontsize'] = '30'
IO_style['penwidth'] = 6

styles = {'IO': {'ro': IO_style.copy(),
                 'rw': IO_style.copy(),
                 'raw': IO_style.copy()
                 },
          'main': default_style.copy(),
          'ext': default_style.copy()}

styles['IO']['ro']['color'] = '#20B2AA '
styles['IO']['rw']['color'] = '#4169E1 '
styles['IO']['raw']['color'] = '#008080 '
styles['ext']['shape'] = 'circle'
styles['ext']['fillcolor'] = 'red'
styles['ext']['style'] = 'filled'


# styles['ext']['fixedsize'] = 'True'


def generate_diagram(filename, plot_extensions=False):
    dia = pygraphviz.AGraph(strict=False, splines='true')
    G = dia
    G.node_attr['fontname'] = 'Arial'
    # G.node_attr['shape'] = 'circle'
    # G.node_attr['fixedsize'] = 'true'
    # G.node_attr['sep'] = '-100'
    G.node_attr['fixedsize'] = 'False'
    # G.graph_attr['overlap'] = 'False'
    G.graph_attr['packMode'] = 'clust'
    # G.graph_attr['levelsgap'] = -500
    G.node_attr['fontsize'] = '20'
    G.edge_attr['minlen'] = '0'
    # G.node_attr['style'] = 'filled'
    # G.graph_attr['outputorder'] = 'edgesfirst'
    # G.graph_attr['splines'] = "compound"
    G.graph_attr['label'] = "NEO {}".format(neo.__version__)
    G.graph_attr['ratio'] = '1.0'
    # G.edge_attr['color'] = '#1100FF'

    G.edge_attr['style'] = 'setlinewidth(4)'

    dia.add_node('NEO', shape='circle', fontsize=50)

    for io in neo.io.iolist:
        io_name = io.name
        rawio = False
        if issubclass(io, neo.io.basefromrawio.BaseFromRaw):
            rawio = True
            if io_name == 'BaseIO':
                io_name = io.__name__.rstrip('RawIO')
        if io_name is None:
            io_name = io.__name__.rstrip('IO')
        if 'example' in io_name:
            continue

        if io.is_writable and io.is_readable:
            mode = 'rw'
        elif io.is_readable:
            mode = 'ro'
        if rawio:
            mode = 'raw'

        dia.add_node(io_name, **styles['IO'][mode])
        dia.add_edge('NEO', io_name)

        if plot_extensions:
            for ext in io.extensions:
                dia.add_node(ext, **styles['ext'])
                dia.add_edge(io_name, ext, minlen=0)

    dia.layout(prog='fdp')  # neato, dot, twopi, circo, fdp, nop, wc, acyclic, gvpr, gvcolor,
    # ccomps, sccmap, tred, sfdp.
    for ext in ['png', 'svg', 'eps']:
        dia.draw('{}.{}'.format(filename, ext))


if __name__ == '__main__':
    generate_diagram('IODiagram', plot_extensions=False)
    generate_diagram('IODiagram_ext', plot_extensions=True)
    # pyplot.show()



'''
Generate datasets for testing
'''

from datetime import datetime
import random
import string
import numpy as np
from numpy.random import rand
import quantities as pq

from neo.core import (AnalogSignal, Block, Epoch, Event, IrregularlySampledSignal, Group,
                      Segment, SpikeTrain, ImageSequence, ChannelView,
                      CircularRegionOfInterest, RectangularRegionOfInterest,
                      PolygonRegionOfInterest)

TEST_ANNOTATIONS = [1, 0, 1.5, "this is a test", datetime.fromtimestamp(424242424), None]


def random_string(length=10):
    return "".join(random.choice(string.ascii_letters) for i in range(length))


def random_datetime(min_year=1990, max_year=datetime.now().year):
    start = datetime(min_year, 1, 1, 0, 0, 0)
    end = datetime(max_year, 12, 31, 23, 59, 59)
    return start + (end - start) * random.random()


def random_annotations(n=1):
    annotation_generators = (
        random.random,
        random_datetime,
        random_string,
        lambda: None
    )
    annotations = {}
    for i in range(n):
        var_name = random_string(6)
        annotation_generator = random.choice(annotation_generators)
        annotations[var_name] = annotation_generator()
    return annotations


def random_signal(name=None, **annotations):
    n_channels = random.randint(1, 7)
    sig_length = random.randint(20, 200)
    if len(annotations) == 0:
        annotations = random_annotations(5)
    obj = AnalogSignal(
        np.random.uniform(size=(sig_length, n_channels)),
        units=random.choice(("mV", "nA")),
        t_start=random.uniform(0, 10) * pq.ms,
        sampling_rate=random.uniform(0.1, 10) * pq.kHz,
        name=name or random_string(),
        file_origin=random_string(),
        description=random_string(100),
        array_annotations=None,   # todo
        **annotations
    )
    return obj


def random_irreg_signal(name=None, **annotations):
    n_channels = random.randint(1, 7)
    sig_length = random.randint(20, 200)
    if len(annotations) == 0:
        annotations = random_annotations(5)
    mean_firing_rate = np.random.uniform(0.1, 10) * pq.kHz
    times = np.cumsum(np.random.uniform(1.0 / mean_firing_rate, size=(sig_length,))) * pq.ms
    obj = IrregularlySampledSignal(
        times,
        np.random.uniform(size=(sig_length, n_channels)),
        units=random.choice(("mV", "nA")),
        name=name or random_string(),
        file_origin=random_string(),
        description=random_string(100),
        array_annotations=None,   # todo
        **annotations
    )
    return obj


def random_event(name=None, **annotations):
    size = random.randint(1, 7)
    times = np.cumsum(np.random.uniform(5, 10, size=size))
    labels = [random_string() for i in range(size)]
    if len(annotations) == 0:
        annotations = random_annotations(3)
    obj = Event(
        times=times,
        labels=labels,
        units="ms",
        name=name or random_string(),
        array_annotations=None,   # todo
        **annotations
    )
    return obj


def random_epoch():
    size = random.randint(1, 7)
    times = np.cumsum(np.random.uniform(5, 10, size=size))
    durations = np.random.uniform(1, 3, size=size)
    labels = [random_string() for i in range(size)]
    obj = Epoch(
        times=times,
        durations=durations,
        labels=labels,
        units="ms",
        name=random_string(),
        array_annotations=None,   # todo
        **random_annotations(3)
    )
    return obj


def random_spiketrain(name=None, **annotations):
    size = random.randint(1, 50)
    times = np.cumsum(np.random.uniform(0.5, 10, size=size))
    if len(annotations) == 0:
        annotations = random_annotations(3)
    # todo: waveforms
    obj = SpikeTrain(
        times=times,
        t_stop=times[-1] + random.uniform(0, 5),
        units="ms",
        name=name or random_string(),
        array_annotations=None,   # todo
        **annotations
    )
    return obj


def random_segment():
    seg = Segment(
        name=random_string(10),
        description=random_string(100),
        file_origin=random_string(20),
        file_datetime=random_datetime(),
        rec_datetime=random_datetime(),
        **random_annotations(4)
    )
    n_sigs = random.randint(0, 5)
    for i in range(n_sigs):
        seg.analogsignals.append(random_signal())
    n_irrsigs = random.randint(0, 5)
    for i in range(n_irrsigs):
        seg.irregularlysampledsignals.append(random_irreg_signal())
    n_events = random.randint(0, 3)
    for i in range(n_events):
        seg.events.append(random_event())
    n_epochs = random.randint(0, 3)
    for i in range(n_epochs):
        seg.epochs.append(random_epoch())
    n_spiketrains = random.randint(0, 20)
    for i in range(n_spiketrains):
        seg.spiketrains.append(random_spiketrain())
    # todo: add some ImageSequence and ROI objects

    for child in seg.data_children:
        child.segment = seg
    return seg


def random_group(candidates):
    if len(candidates) == 0:
        return None
    elif len(candidates) == 1:
        objects = candidates
    else:
        k = random.randint(1, len(candidates))
        objects = random.sample(candidates, k)
    obj = Group(objects=objects,
                name=random_string(),
                **random_annotations(5))
    return obj


def random_channelview(signal):
    n_channels = signal.shape[1]
    if n_channels > 2:
        view_size = np.random.randint(1, n_channels - 1)
        index = np.random.choice(np.arange(signal.shape[1]), view_size, replace=False)
        obj = ChannelView(
            signal,
            index,
            name=random_string(),
            **random_annotations(3)
        )
        return obj
    else:
        return None


def random_block():
    block = Block(
        name=random_string(10),
        description=random_string(100),
        file_origin=random_string(20),
        file_datetime=random_datetime(),
        rec_datetime=random_datetime(),
        **random_annotations(6)
    )
    n_seg = random.randint(0, 5)
    for i in range(n_seg):
        seg = random_segment()
        block.segments.append(seg)
        seg.block = block
    children = list(block.data_children_recur)
    views = []
    for child in children:
        if isinstance(child, (AnalogSignal, IrregularlySampledSignal)):
            PROB_SIGNAL_HAS_VIEW = 0.5
            if np.random.random_sample() < PROB_SIGNAL_HAS_VIEW:
                chv = random_channelview(child)
                if chv:
                    views.append(chv)
    children.extend(views)
    n_groups = random.randint(0, 5)
    for i in range(n_groups):
        group = random_group(children)
        if group:
            block.groups.append(group)
            group.block = block
            children.append(group)  # this can give us nested groups
    return block


def simple_block():
    block = Block(
        name="test block",
        species="rat",
        brain_region="cortex"
    )
    block.segments = [
        Segment(name="test segment #1",
                cell_type="spiny stellate"),
        Segment(name="test segment #2",
                cell_type="pyramidal",
                thing="amajig")
    ]
    for segment in block.segments:
        segment.block = block
    block.segments[0].analogsignals.extend((
        random_signal(name="signal #1 in segment #1", thing="wotsit"),
        random_signal(name="signal #2 in segment #1", thing="frooble"),
    ))
    block.segments[1].analogsignals.extend((
        random_signal(name="signal #1 in segment #2", thing="amajig"),
    ))
    block.segments[1].irregularlysampledsignals.extend((
        random_irreg_signal(name="signal #1 in segment #2", thing="amajig"),
    ))
    block.segments[0].events.extend((
        random_event(name="event array #1 in segment #1", thing="frooble"),
    ))
    block.segments[1].events.extend((
        random_event(name="event array #1 in segment #2", thing="wotsit"),
    ))
    block.segments[0].spiketrains.extend((
        random_spiketrain(name="spiketrain #1 in segment #1", thing="frooble"),
        random_spiketrain(name="spiketrain #2 in segment #1", thing="wotsit")
    ))
    return block


def generate_one_simple_block(block_name='block_0', nb_segment=3, supported_objects=[], **kws):
    if supported_objects and Block not in supported_objects:
        raise ValueError('Block must be in supported_objects')
    bl = Block()  # name = block_name)

    objects = supported_objects
    if Segment in objects:
        for s in range(nb_segment):
            seg = generate_one_simple_segment(seg_name="seg" + str(s), supported_objects=objects,
                                              **kws)
            bl.segments.append(seg)

    # if RecordingChannel in objects:
    #    populate_RecordingChannel(bl)

    bl.create_many_to_one_relationship()
    return bl


def generate_one_simple_segment(seg_name='segment 0', supported_objects=[], nb_analogsignal=4,
                                t_start=0. * pq.s, sampling_rate=10 * pq.kHz, duration=6. * pq.s,

                                nb_spiketrain=6, spikerate_range=[.5 * pq.Hz, 12 * pq.Hz],

                                event_types={'stim': ['a', 'b', 'c', 'd'],
                                             'enter_zone': ['one', 'two'],
                                             'color': ['black', 'yellow', 'green'], },
                                event_size_range=[5, 20],

                                epoch_types={'animal state': ['Sleep', 'Freeze', 'Escape'],
                                             'light': ['dark', 'lighted']},
                                epoch_duration_range=[.5, 3.],
                                # this should be multiplied by pq.s, no?

                                array_annotations={'valid': np.array([True, False]),
                                                   'number': np.array(range(5))}

                                ):
    if supported_objects and Segment not in supported_objects:
        raise ValueError('Segment must be in supported_objects')
    seg = Segment(name=seg_name)
    if AnalogSignal in supported_objects:
        for a in range(nb_analogsignal):
            anasig = AnalogSignal(rand(int((sampling_rate * duration).simplified)),
                                  sampling_rate=sampling_rate,
                                  t_start=t_start, units=pq.mV, channel_index=a,
                                  name='sig %d for segment %s' % (a, seg.name))
            seg.analogsignals.append(anasig)

    if SpikeTrain in supported_objects:
        for s in range(nb_spiketrain):
            spikerate = rand() * np.diff(spikerate_range)
            spikerate += spikerate_range[0].magnitude
            # spikedata = rand(int((spikerate*duration).simplified))*duration
            # sptr = SpikeTrain(spikedata,
            #                  t_start=t_start, t_stop=t_start+duration)
            #                  #, name = 'spiketrain %d'%s)
            spikes = rand(int((spikerate * duration).simplified))
            spikes.sort()  # spikes are supposed to be an ascending sequence
            sptr = SpikeTrain(spikes * duration, t_start=t_start, t_stop=t_start + duration)
            sptr.annotations['channel_index'] = s
            # Randomly generate array_annotations from given options
            arr_ann = {key: value[(rand(len(spikes)) * len(value)).astype('i')] for (key, value) in
                       array_annotations.items()}
            sptr.array_annotate(**arr_ann)
            seg.spiketrains.append(sptr)

    if Event in supported_objects:
        for name, labels in event_types.items():
            evt_size = rand() * np.diff(event_size_range)
            evt_size += event_size_range[0]
            evt_size = int(evt_size)
            labels = np.array(labels, dtype='U')
            labels = labels[(rand(evt_size) * len(labels)).astype('i')]
            evt = Event(times=rand(evt_size) * duration, labels=labels)
            # Randomly generate array_annotations from given options
            arr_ann = {key: value[(rand(evt_size) * len(value)).astype('i')] for (key, value) in
                       array_annotations.items()}
            evt.array_annotate(**arr_ann)
            seg.events.append(evt)

    if Epoch in supported_objects:
        for name, labels in epoch_types.items():
            t = 0
            times = []
            durations = []
            while t < duration:
                times.append(t)
                dur = rand() * (epoch_duration_range[1] - epoch_duration_range[0])
                dur += epoch_duration_range[0]
                durations.append(dur)
                t = t + dur
            labels = np.array(labels, dtype='U')
            labels = labels[(rand(len(times)) * len(labels)).astype('i')]
            assert len(times) == len(durations)
            assert len(times) == len(labels)
            epc = Epoch(times=pq.Quantity(times, units=pq.s),
                        durations=pq.Quantity(durations, units=pq.s),
                        labels=labels,)
            assert epc.times.dtype == 'float'
            # Randomly generate array_annotations from given options
            arr_ann = {key: value[(rand(len(times)) * len(value)).astype('i')] for (key, value) in
                       array_annotations.items()}
            epc.array_annotate(**arr_ann)
            seg.epochs.append(epc)

    # TODO : Spike, Event

    seg.create_many_to_one_relationship()
    return seg


def generate_from_supported_objects(supported_objects):
    # ~ create_many_to_one_relationship
    if not supported_objects:
        raise ValueError('No objects specified')
    objects = supported_objects
    if Block in supported_objects:
        higher = generate_one_simple_block(supported_objects=objects)
    elif Segment in objects:
        higher = generate_one_simple_segment(supported_objects=objects)
    else:
        # TODO
        return None

    higher.create_many_to_one_relationship()
    return higher



"""
Class for reading data from WinWCP, a software tool written by
John Dempster.

WinWCP is free:
http://spider.science.strath.ac.uk/sipbs/software.htm

Author: Samuel Garcia
"""
from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                _spike_channel_dtype, _event_channel_dtype, _common_sig_characteristics)

import numpy as np

import struct


class WinWcpRawIO(BaseRawIO):
    extensions = ['wcp']
    rawmode = 'one-file'

    def __init__(self, filename=''):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        SECTORSIZE = 512

        # only one memmap for all segment to avoid
        # "error: [Errno 24] Too many open files"
        self._memmap = np.memmap(self.filename, dtype='uint8', mode='r')

        with open(self.filename, 'rb') as fid:

            headertext = fid.read(1024)
            headertext = headertext.decode('ascii')
            header = {}
            for line in headertext.split('\r\n'):
                if '=' not in line:
                    continue
                # print '#' , line , '#'
                key, val = line.split('=')
                if key in ['NC', 'NR', 'NBH', 'NBA', 'NBD', 'ADCMAX', 'NP', 'NZ', ]:
                    val = int(val)
                elif key in ['AD', 'DT', ]:
                    val = val.replace(',', '.')
                    val = float(val)
                header[key] = val

            nb_segment = header['NR']
            self._raw_signals = {}
            all_sampling_interval = []
            # loop for record number
            for seg_index in range(header['NR']):
                offset = 1024 + seg_index * (SECTORSIZE * header['NBD'] + 1024)

                # read analysis zone
                analysisHeader = HeaderReader(fid, AnalysisDescription).read_f(offset=offset)

                # read data
                NP = (SECTORSIZE * header['NBD']) // 2
                NP = NP - NP % header['NC']
                NP = NP // header['NC']
                NC = header['NC']
                ind0 = offset + header['NBA'] * SECTORSIZE
                ind1 = ind0 + NP * NC * 2
                sigs = self._memmap[ind0:ind1].view('int16').reshape(NP, NC)
                self._raw_signals[seg_index] = sigs

                all_sampling_interval.append(analysisHeader['SamplingInterval'])

        # sampling interval can be slightly varying due to float precision
        # all_sampling_interval are not always unique
        self._sampling_rate = 1. / np.median(all_sampling_interval)

        signal_channels = []
        for c in range(header['NC']):
            YG = float(header['YG%d' % c].replace(',', '.'))
            ADCMAX = header['ADCMAX']
            VMax = analysisHeader['VMax'][c]

            name = header['YN%d' % c]
            chan_id = header['YO%d' % c]
            units = header['YU%d' % c]
            gain = VMax / ADCMAX / YG
            offset = 0.
            stream_id = '0'
            signal_channels.append((name, chan_id, self._sampling_rate, 'int16',
                                 units, gain, offset, stream_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        characteristics = signal_channels[_common_sig_characteristics]
        unique_characteristics = np.unique(characteristics)
        signal_streams = []
        for i in range(unique_characteristics.size):
            mask = unique_characteristics[i] == characteristics
            signal_channels['stream_id'][mask] = str(i)
            signal_streams.append((f'stream {i}', str(i)))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [nb_segment]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return 0.

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals[seg_index].shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._raw_signals[seg_index].shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        stream_id = self.header['signal_streams'][stream_index]['id']
        global_channel_indexes, = np.nonzero(self.header['signal_channels']
                                    ['stream_id'] == stream_id)
        if channel_indexes is None:
            channel_indexes = slice(None)
        inds = global_channel_indexes[channel_indexes]
        raw_signals = self._raw_signals[seg_index][slice(i_start, i_stop), inds]
        return raw_signals


AnalysisDescription = [
    ('RecordStatus', '8s'),
    ('RecordType', '4s'),
    ('GroupNumber', 'f'),
    ('TimeRecorded', 'f'),
    ('SamplingInterval', 'f'),
    ('VMax', '8f'),
]


class HeaderReader():
    def __init__(self, fid, description):
        self.fid = fid
        self.description = description

    def read_f(self, offset=0):
        self.fid.seek(offset)
        d = {}
        for key, fmt in self.description:
            val = struct.unpack(fmt, self.fid.read(struct.calcsize(fmt)))
            if len(val) == 1:
                val = val[0]
            else:
                val = list(val)
            d[key] = val
        return d



"""
PhyRawIO is a class to handle Phy spike sorting data.
Ported from:
https://github.com/SpikeInterface/spikeextractors/blob/
f20b1219eba9d3330d5d7cd7ce8d8924a255b8c2/spikeextractors/
extractors/phyextractors/phyextractors.py

Author: Regimantas Jurkus
"""

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                        _spike_channel_dtype, _event_channel_dtype)

import numpy as np
from pathlib import Path
import re
import csv
import ast


class PhyRawIO(BaseRawIO):
    """
    Class for reading Phy data.

    Usage:
        >>> import neo.rawio
        >>> r = neo.rawio.PhyRawIO(dirname='/dir/to/phy/folder')
        >>> r.parse_header()
        >>> print(r)
        >>> spike_timestamp = r.get_spike_timestamps(block_index=0,
        ... seg_index=0, spike_channel_index=0, t_start=None, t_stop=None)
        >>> spike_times = r.rescale_spike_timestamp(spike_timestamp, 'float64')

    """
    extensions = []
    rawmode = 'one-dir'

    def __init__(self, dirname=''):
        BaseRawIO.__init__(self)
        self.dirname = dirname

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        phy_folder = Path(self.dirname)

        self._spike_times = np.load(phy_folder / 'spike_times.npy')
        self._spike_templates = np.load(phy_folder / 'spike_templates.npy')

        if (phy_folder / 'spike_clusters.npy').is_file():
            self._spike_clusters = np.load(phy_folder / 'spike_clusters.npy')
        else:
            self._spike_clusters = self._spike_templates

        # TODO: Add this when array_annotations are ready
        # if (phy_folder / 'amplitudes.npy').is_file():
        #     amplitudes = np.squeeze(np.load(phy_folder / 'amplitudes.npy'))
        # else:
        #     amplitudes = np.ones(len(spike_times))
        #
        # if (phy_folder / 'pc_features.npy').is_file():
        #     pc_features = np.squeeze(np.load(phy_folder / 'pc_features.npy'))
        # else:
        #     pc_features = None

        # SEE: https://stackoverflow.com/questions/4388626/
        #  python-safe-eval-string-to-bool-int-float-none-string
        if (phy_folder / 'params.py').is_file():
            with (phy_folder / 'params.py').open('r') as f:
                contents = f.read()
            metadata = dict()
            contents = contents.replace('\n', ' ')
            pattern = re.compile(r'(\S*)[\s]?=[\s]?(\S*)')
            elements = pattern.findall(contents)
            for key, value in elements:
                metadata[key.lower()] = ast.literal_eval(value)

        self._sampling_frequency = metadata['sample_rate']

        clust_ids = np.unique(self._spike_clusters)
        self.unit_labels = list(clust_ids)

        self._t_start = 0.
        self._t_stop = max(self._spike_times).item() / self._sampling_frequency

        signal_streams = []
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        signal_channels = []
        signal_channels = np.array(signal_channels,
                                   dtype=_signal_channel_dtype)

        spike_channels = []
        for i, clust_id in enumerate(clust_ids):
            unit_name = f'unit {clust_id}'
            unit_id = f'{clust_id}'
            wf_units = ''
            wf_gain = 0
            wf_offset = 0.
            wf_left_sweep = 0
            wf_sampling_rate = 0
            spike_channels.append((unit_name, unit_id, wf_units, wf_gain,
                                  wf_offset, wf_left_sweep, wf_sampling_rate))
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [1]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        self._generate_minimal_annotations()

        csv_tsv_files = [x for x in phy_folder.iterdir() if
                         x.suffix == '.csv' or x.suffix == '.tsv']

        # annotation_lists is list of list of dict (python==3.8)
        # or list of list of ordered dict (python==3.6)
        # SEE: https://docs.python.org/3/library/csv.html#csv.DictReader
        annotation_lists = [self._parse_tsv_or_csv_to_list_of_dict(file)
                            for file in csv_tsv_files]

        bl_ann = self.raw_annotations['blocks'][0]
        bl_ann['name'] = "Block #0"
        seg_ann = bl_ann['segments'][0]
        seg_ann['name'] = 'Seg #0 Block #0'
        for index, clust_id in enumerate(clust_ids):
            spiketrain_an = seg_ann['spikes'][index]

            # Add cluster_id annotation
            spiketrain_an['cluster_id'] = clust_id

            # Loop over list of list of dict and annotate each st
            for annotation_list in annotation_lists:
                clust_key, property_name = tuple(annotation_list[0].
                                                 keys())
                if property_name == 'KSLabel':
                    annotation_name = 'quality'
                else:
                    annotation_name = property_name.lower()
                for annotation_dict in annotation_list:
                    if int(annotation_dict[clust_key]) == clust_id:
                        spiketrain_an[annotation_name] = \
                            annotation_dict[property_name]
                        break

    def _segment_t_start(self, block_index, seg_index):
        assert block_index == 0
        return self._t_start

    def _segment_t_stop(self, block_index, seg_index):
        assert block_index == 0
        return self._t_stop

    def _get_signal_size(self, block_index, seg_index, channel_indexes=None):
        return None

    def _get_signal_t_start(self, block_index, seg_index, channel_indexes):
        return None

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                channel_indexes):
        return None

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        assert block_index == 0
        spikes = self._spike_clusters
        unit_label = self.unit_labels[spike_channel_index]
        mask = spikes == unit_label
        nb_spikes = np.sum(mask)
        return nb_spikes

    def _get_spike_timestamps(self, block_index, seg_index,
                              spike_channel_index, t_start, t_stop):
        assert block_index == 0
        assert seg_index == 0

        unit_label = self.unit_labels[spike_channel_index]
        mask = self._spike_clusters == unit_label
        spike_timestamps = self._spike_times[mask]

        if t_start is not None:
            start_frame = int(t_start * self._sampling_frequency)
            spike_timestamps = \
                spike_timestamps[spike_timestamps >= start_frame]
        if t_stop is not None:
            end_frame = int(t_stop * self._sampling_frequency)
            spike_timestamps = spike_timestamps[spike_timestamps < end_frame]

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self._sampling_frequency
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index,
                                 spike_channel_index, t_start, t_stop):
        return None

    def _event_count(self, block_index, seg_index, event_channel_index):
        return None

    def _get_event_timestamps(self, block_index, seg_index,
                              event_channel_index, t_start, t_stop):
        return None

    def _rescale_event_timestamp(self, event_timestamps, dtype):
        return None

    def _rescale_epoch_duration(self, raw_duration, dtype):
        return None

    @staticmethod
    def _parse_tsv_or_csv_to_list_of_dict(filename):
        list_of_dict = list()
        letter_pattern = re.compile('[a-zA-Z]')
        float_pattern = re.compile(r'\d*\.')
        with open(filename) as csvfile:
            if filename.suffix == '.csv':
                reader = csv.DictReader(csvfile, delimiter=',')
            elif filename.suffix == '.tsv':
                reader = csv.DictReader(csvfile, delimiter='\t')
            else:
                raise ValueError("Function parses only .csv or .tsv files")
            line = 0

            for row in reader:
                if line == 0:
                    key1, key2 = tuple(row.keys())
                # Convert cluster ID to int
                row[key1] = int(row[key1])
                # Convert strings without letters
                if letter_pattern.match(row[key2]) is None:
                    if float_pattern.match(row[key2]) is None:
                        row[key2] = int(row[key2])
                    else:
                        row[key2] = float(row[key2])

                list_of_dict.append(row)
                line += 1

        return list_of_dict



# -*- coding: utf-8 -*-
"""
This module implements :class:`SpikeTrainList`, a pseudo-list
which supports a multiplexed representation of spike trains
(all times in a single array, with a second array indicating which
neuron/channel the spike is from).
"""

import numpy as np
import quantities as pq
from .spiketrain import SpikeTrain, normalize_times_array


def is_spiketrain_or_proxy(obj):
    return isinstance(obj, SpikeTrain) or getattr(obj, "proxy_for", None) == SpikeTrain


class SpikeTrainList(object):
    """
    This class contains multiple spike trains, and can represent them
    either as a list of SpikeTrain objects or as a pair of arrays
    (all spike times in a single array, with a second array indicating which
    neuron/channel the spike is from).

    A SpikeTrainList object should behave like a list of SpikeTrains
    for iteration and item access. It is not intended to be used directly
    by users, but is available as the attribute `spiketrains` of Segments.

    Examples:

        # Create from list of SpikeTrain objects

        >>> stl = SpikeTrainList(items=(
        ...     SpikeTrain([0.5, 0.6, 23.6, 99.2], units="ms", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
        ...     SpikeTrain([0.0007, 0.0112], units="s", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
        ...     SpikeTrain([1100, 88500], units="us", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
        ...     SpikeTrain([], units="ms", t_start=0 * pq.ms, t_stop=100.0 * pq.ms),
        ... ))
        >>> stl.multiplexed
        (array([0, 0, 0, 0, 1, 1, 2, 2]),
         array([ 0.5,  0.6, 23.6, 99.2,  0.7, 11.2,  1.1, 88.5]) * ms)

        # Create from a pair of arrays

        >>> stl = SpikeTrainList.from_spike_time_array(
        ...     np.array([0.5, 0.6, 0.7, 1.1, 11.2, 23.6, 88.5, 99.2]),
        ...     np.array([0, 0, 1, 2, 1, 0, 2, 0]),
        ...     all_channel_ids=[0, 1, 2, 3],
        ...     units='ms',
        ...     t_start=0 * pq.ms,
        ...     t_stop=100.0 * pq.ms)
        >>> list(stl)
        [<SpikeTrain(array([ 0.5,  0.6, 23.6, 99.2]) * ms, [0.0 ms, 100.0 ms])>,
         <SpikeTrain(array([ 0.7, 11.2]) * ms, [0.0 ms, 100.0 ms])>,
         <SpikeTrain(array([ 1.1, 88.5]) * ms, [0.0 ms, 100.0 ms])>,
         <SpikeTrain(array([], dtype=float64) * ms, [0.0 ms, 100.0 ms])>]

    """

    def __init__(self, items=None, segment=None):
        """Initialize self"""
        if items is None:
            self._items = items
        else:
            for item in items:
                if not is_spiketrain_or_proxy(item):
                    raise ValueError(
                        "`items` can only contain SpikeTrain objects or proxy pbjects")
            self._items = list(items)
        self._spike_time_array = None
        self._channel_id_array = None
        self._all_channel_ids = None
        self._spiketrain_metadata = None
        self.segment = segment

    def __iter__(self):
        """Implement iter(self)"""
        if self._items is None:
            self._spiketrains_from_array()
        for item in self._items:
            yield item

    def __getitem__(self, i):
        """x.__getitem__(y) <==> x[y]"""
        if self._items is None:
            self._spiketrains_from_array()
        items = self._items[i]
        if is_spiketrain_or_proxy(items):
            return items
        else:
            return SpikeTrainList(items=items)

    def __str__(self):
        """Return str(self)"""
        if self._items is None:
            if self._spike_time_array is None:
                return str([])
            else:
                return "SpikeTrainList containing {} spikes from {} neurons".format(
                    self._spike_time_array.size,
                    len(self._all_channel_ids))
        else:
            return str(self._items)

    def __len__(self):
        """Return len(self)"""
        if self._items is None:
            if self._all_channel_ids is not None:
                return len(self._all_channel_ids)
            elif self._channel_id_array is not None:
                return np.unique(self._channel_id_array).size
            else:
                return 0
        else:
            return len(self._items)

    def _add_spiketrainlists(self, other, in_place=False):
        if self._spike_time_array is None or other._spike_time_array is None:
            # if either self or other is not storing multiplexed spike trains
            # we combine them using the list of SpikeTrains representation
            if self._items is None:
                self._spiketrains_from_array()
            if other._items is None:
                other._spiketrains_from_array()
            if in_place:
                self._items.extend(other._items)
                return self
            else:
                return self.__class__(items=self._items[:] + other._items)
        else:
            # both self and other are storing multiplexed spike trains
            # so we update the array representation
            if self._spiketrain_metadata['t_start'] != other._spiketrain_metadata['t_start']:
                raise ValueError("Incompatible t_start")
                # todo: adjust times and t_start of other to be compatible with self
            if self._spiketrain_metadata['t_stop'] != other._spiketrain_metadata['t_stop']:
                raise ValueError("Incompatible t_stop")
                # todo: adjust t_stop of self and other as necessary
            combined_spike_time_array = np.hstack(
                (self._spike_time_array, other._spike_time_array))
            combined_channel_id_array = np.hstack(
                (self._channel_id_array, other._channel_id_array))
            combined_channel_ids = set(list(self._all_channel_ids) + other._all_channel_ids)
            if len(combined_channel_ids) != (
                len(self._all_channel_ids) + len(other._all_channel_ids)
            ):
                raise ValueError("Duplicate channel ids, please rename channels before adding")
            if in_place:
                self._spike_time_array = combined_spike_time_array
                self._channel_id_array = combined_channel_id_array
                self._all_channel_ids = combined_channel_ids
                self._items = None
                return self
            else:
                return self.__class__.from_spike_time_array(
                    combined_spike_time_array,
                    combined_channel_id_array,
                    combined_channel_ids,
                    t_start=self._spiketrain_metadata['t_start'],
                    t_stop=self._spiketrain_metadata['t_stop'])

    def __add__(self, other):
        """Return self + other"""
        if isinstance(other, self.__class__):
            return self._add_spiketrainlists(other)
        elif other and is_spiketrain_or_proxy(other[0]):
            return self._add_spiketrainlists(
                self.__class__(items=other, segment=self.segment)
            )
        else:
            if self._items is None:
                self._spiketrains_from_array()
            return self._items + other

    def __iadd__(self, other):
        """Return self"""
        if isinstance(other, self.__class__):
            return self._add_spiketrainlists(other, in_place=True)
        elif other and is_spiketrain_or_proxy(other[0]):
            for obj in other:
                obj.segment = self.segment
            if self._items is None:
                self._spiketrains_from_array()
            self._items.extend(other)
            return self
        else:
            raise TypeError("Can only add a SpikeTrainList or a list of SpikeTrains in place")

    def __radd__(self, other):
        """Return other + self"""
        if isinstance(other, self.__class__):
            return other._add_spiketrainlists(self)
        elif other and is_spiketrain_or_proxy(other[0]):
            for obj in other:
                obj.segment = self.segment
            if self._items is None:
                self._spiketrains_from_array()
            self._items.extend(other)
            return self
        elif len(other) == 0:
            return self
        else:
            if self._items is None:
                self._spiketrains_from_array()
            return other + self._items

    def append(self, obj):
        """L.append(object) -> None -- append object to end"""
        if not is_spiketrain_or_proxy(obj):
            raise ValueError("Can only append SpikeTrain objects")
        if self._items is None:
            self._spiketrains_from_array()
        obj.segment = self.segment
        self._items.append(obj)

    def extend(self, iterable):
        """L.extend(iterable) -> None -- extend list by appending elements from the iterable"""
        if self._items is None:
            self._spiketrains_from_array()
        for obj in iterable:
            obj.segment = self.segment
        self._items.extend(iterable)

    @classmethod
    def from_spike_time_array(cls, spike_time_array, channel_id_array,
                              all_channel_ids, t_stop, units=None,
                              t_start=0.0 * pq.s, **annotations):
        """Create a SpikeTrainList object from an array of spike times
        and an array of channel ids.

        *Required attributes/properties*:

        :spike_time_array: (quantity array 1D, numpy array 1D, or list) The times of
            all spikes.
        :channel_id_array: (numpy array 1D of dtype int) The id of the channel (e.g. the
            neuron) to which each spike belongs. This array should have the same length
            as :attr:`spike_time_array`
        :all_channel_ids: (list, tuple, or numpy array 1D containing integers) All
            channel ids. This is needed to represent channels in which there are no
            spikes.
        :units: (quantity units) Required if :attr:`spike_time_array` is not a
                :class:`~quantities.Quantity`.
        :t_stop: (quantity scalar, numpy scalar, or float) Time at which
            spike recording ended. This will be converted to the
            same units as :attr:`spike_time_array` or :attr:`units`.

        *Recommended attributes/properties*:
        :t_start: (quantity scalar, numpy scalar, or float) Time at which
            spike recording began. This will be converted to the
            same units as :attr:`spike_time_array` or :attr:`units`.
            Default: 0.0 seconds.


        *Optional attributes/properties*:
        """
        spike_time_array, dim = normalize_times_array(spike_time_array, units)
        obj = cls()
        obj._spike_time_array = spike_time_array
        obj._channel_id_array = channel_id_array
        obj._all_channel_ids = all_channel_ids
        obj._spiketrain_metadata = {
            "t_start": t_start,
            "t_stop": t_stop
        }
        for name, ann_value in annotations.items():
            if len(ann_value) != len(obj):
                raise ValueError(f"incorrect length for annotation '{name}'")
        obj._annotations = annotations
        return obj

    def _spiketrains_from_array(self):
        """Convert multiplexed spike time data into a list of SpikeTrain objects"""
        if self._spike_time_array is None:
            self._items = []
        else:
            self._items = []
            for i, channel_id in enumerate(self._all_channel_ids):
                mask = self._channel_id_array == channel_id
                times = self._spike_time_array[mask]
                spiketrain = SpikeTrain(times, **self._spiketrain_metadata)
                spiketrain.annotations = {
                    name: value[i]
                    for name, value in self._annotations.items()
                }
                spiketrain.annotate(channel_id=channel_id)
                spiketrain.segment = self.segment
                self._items.append(spiketrain)

    @property
    def multiplexed(self):
        """Return spike trains as a pair of arrays.

        The first (plain NumPy) array contains the ids of the channels/neurons that produced
        each spike, the second (Quantity) array contains the times of the spikes.
        """
        if self._spike_time_array is None:
            # need to convert list of SpikeTrains into multiplexed spike times array
            if self._items is None:
                return np.array([]), np.array([])
            else:
                channel_ids = []
                spike_times = []
                dim = self._items[0].units.dimensionality
                for i, spiketrain in enumerate(self._items):
                    if hasattr(spiketrain, "load"):  # proxy object
                        spiketrain = spiketrain.load()
                    if spiketrain.times.dimensionality.items() == dim.items():
                        # no need to rescale
                        spike_times.append(spiketrain.times)
                    else:
                        spike_times.append(spiketrain.times.rescale(dim))
                    if ("channel_id" in spiketrain.annotations
                        and isinstance(spiketrain.annotations["channel_id"], int)
                        ):
                        ch_id = spiketrain.annotations["channel_id"]
                    else:
                        ch_id = i
                    channel_ids.append(ch_id * np.ones(spiketrain.shape, dtype=np.int64))
                self._spike_time_array = np.hstack(spike_times) * self._items[0].units
                self._channel_id_array = np.hstack(channel_ids)
        return self._channel_id_array, self._spike_time_array



"""
Class for reading data from a SpikeGLX system  (NI-DAQ for neuropixel probe)
See https://billkarsh.github.io/SpikeGLX/


Here an adaptation of the spikeglx tools into the neo rawio API.

Note that each pair of ".bin"/."meta" files is represented as a stream of channels
that share the same sampling rate.
It will be one AnalogSignal multi channel at neo.io level.

Contrary to other implementations this IO reads the entire folder and subfolder and:
  * deals with severals segment based on the `_gt0`, `_gt1`, `_gt2`, etc postfixes
  * deals with all signals "imec0", "imec1" for neuropixel probes and also
     external signal like"nidq". This is the "device"
  * For imec device both "ap" and "lf" are extracted so one device have several "streams"

Note:
  * there are several versions depending the neuropixel probe generation (`1.x`/`2.x`/`3.x`)
    Here, we assume that the `meta` file has the same structure across all generations.
    This need so be checked.
    This IO is developed based on neuropixel generation 2.0, single shank recordings.


# Not implemented yet in this reader:
  * contact SpkeGLX developer to see how to deal with absolut t_start when several segment
  * contact SpkeGLX developer to understand the last channel SY0 function
  * better handling of annotations at object level by sub group of device (after rawio change)
  * better handling of channel location


See:
https://billkarsh.github.io/SpikeGLX/
https://billkarsh.github.io/SpikeGLX/#offline-analysis-tools
https://billkarsh.github.io/SpikeGLX/#metadata-guides
https://github.com/SpikeInterface/spikeextractors/blob/master/spikeextractors/extractors/spikeglxrecordingextractor/spikeglxrecordingextractor.py

This reader handle:

imDatPrb_type=1 (NP 1.0)
imDatPrb_type=21 (NP 2.0, single multiplexed shank)
imDatPrb_type=24 (NP 2.0, 4-shank)

Author : Samuel Garcia
"""

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                _spike_channel_dtype, _event_channel_dtype)

from pathlib import Path
import os
import re

import numpy as np


class SpikeGLXRawIO(BaseRawIO):
    """
    Class for reading data from a SpikeGLX system

    dirname:
        The spikeglx folder containing meta/bin files
    load_sync_channel=False/True
        The last channel (SY0) of each stream is a fake channel used for synchronisation.
    """
    extensions = []
    rawmode = 'one-dir'

    def __init__(self, dirname='', load_sync_channel=False, load_channel_location=False):
        BaseRawIO.__init__(self)
        self.dirname = dirname
        self.load_sync_channel = load_sync_channel
        self.load_channel_location = load_channel_location

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        self.signals_info_list = scan_files(self.dirname)

        # sort stream_name by higher sampling rate first
        srates = {info['stream_name']: info['sampling_rate'] for info in self.signals_info_list}
        stream_names = sorted(list(srates.keys()), key=lambda e: srates[e])[::-1]

        nb_segment = np.unique([info['seg_index'] for info in self.signals_info_list]).size

        self._memmaps = {}
        self.signals_info_dict = {}
        for info in self.signals_info_list:
            # key is (seg_index, stream_name)
            key = (info['seg_index'], info['stream_name'])
            self.signals_info_dict[key] = info

            # create memmap
            data = np.memmap(info['bin_file'], dtype='int16', mode='r', offset=0, order='C')
            # this should be (info['sample_length'], info['num_chan'])
            # be some file are shorten
            data = data.reshape(-1, info['num_chan'])
            self._memmaps[key] = data

        # create channel header
        signal_streams = []
        signal_channels = []
        for stream_name in stream_names:
            # take first segment
            info = self.signals_info_dict[0, stream_name]

            stream_id = stream_name
            stream_index = stream_names.index(info['stream_name'])
            signal_streams.append((stream_name, stream_id))

            # add channels to global list
            for local_chan in range(info['num_chan']):
                chan_name = info['channel_names'][local_chan]
                chan_id = f'{stream_name}#{chan_name}'
                signal_channels.append((chan_name, chan_id, info['sampling_rate'], 'int16',
                                    info['units'], info['channel_gains'][local_chan],
                                    info['channel_offsets'][local_chan], stream_id))
            if not self.load_sync_channel:
                signal_channels = signal_channels[:-1]

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # deal with nb_segment and t_start/t_stop per segment
        self._t_starts = {seg_index: 0. for seg_index in range(nb_segment)}
        self._t_stops = {seg_index: 0. for seg_index in range(nb_segment)}
        for seg_index in range(nb_segment):
            for stream_name in stream_names:
                info = self.signals_info_dict[seg_index, stream_name]
                t_stop = info['sample_length'] / info['sampling_rate']
                self._t_stops[seg_index] = max(self._t_stops[seg_index], t_stop)

        # fille into header dict
        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [nb_segment]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()
        self._generate_minimal_annotations()
        block_ann = self.raw_annotations['blocks'][0]

        for seg_index in range(nb_segment):
            seg_ann = self.raw_annotations['blocks'][0]['segments'][seg_index]
            seg_ann['name'] = "Segment {}".format(seg_index)

            for c, signal_stream in enumerate(signal_streams):
                stream_name = signal_stream['name']
                sig_ann = self.raw_annotations['blocks'][0]['segments'][seg_index]['signals'][c]

                if self.load_channel_location:
                    # need probeinterface to be installed
                    import probeinterface
                    info = self.signals_info_dict[seg_index, stream_name]
                    if 'imroTbl' in info['meta'] and info['signal_kind'] == 'ap':
                        # only for ap channel
                        probe = probeinterface.read_spikeglx(info['meta_file'])
                        loc = probe.contact_positions
                        if self.load_sync_channel:
                            # one fake channel  for "sys0"
                            loc = np.concatenate((loc, [[0., 0.]]), axis=0)
                        for ndim in range(loc.shape[1]):
                            sig_ann['__array_annotations__'][f'channel_location_{ndim}'] = loc[:, ndim]

    def _segment_t_start(self, block_index, seg_index):
        return 0.

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stops[seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header['signal_streams'][stream_index]['id']
        memmap = self._memmaps[seg_index, stream_id]
        return int(memmap.shape[0])

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        stream_id = self.header['signal_streams'][stream_index]['id']
        memmap = self._memmaps[seg_index, stream_id]
        if channel_indexes is None:
            if self.load_sync_channel:
                channel_selection = slice(None)
            else:
                channel_selection = slice(-1)
        elif isinstance(channel_indexes, slice):
            if self.load_sync_channel:
                # simple
                channel_selection = channel_indexes
            else:
                # more tricky because negative
                sl_start = channel_indexes.start
                sl_stop = channel_indexes.stop
                sl_step = channel_indexes.step
                if sl_stop is not None and sl_stop < 0:
                    sl_stop = sl_stop - 1
                elif sl_stop is None:
                    sl_stop = -1
                channel_selection = slice(sl_start, sl_stop, sl_step)
        elif not isinstance(channel_indexes, slice):
            if np.all(np.diff(channel_indexes) == 1):
                # consecutive channel then slice this avoid a copy (because of ndarray.take(...)
                # and so keep the underlying memmap
                channel_selection = slice(channel_indexes[0],
                                          channel_indexes[0] + len(channel_indexes))
            else:
                channel_selection = channel_indexes
        else:
            raise ValueError('get_analogsignal_chunk : channel_indexes'
                             'must be slice or list or array of int')

        raw_signals = memmap[slice(i_start, i_stop), channel_selection]

        return raw_signals


def scan_files(dirname):
    """
    Scan for pairs of `.bin` and `.meta` files and return information about it.
    """
    info_list = []

    for root, dirs, files in os.walk(dirname):

        for file in files:
            if not file.endswith('.meta'):
                continue
            meta_filename = Path(root) / file
            bin_filename = Path(root) / file.replace('.meta', '.bin')
            if meta_filename.exists() and bin_filename.exists():
                meta = read_meta_file(meta_filename)
            else:
                continue

            num_chan = int(meta['nSavedChans'])

            # Example file name structure:
            # Consider the filenames: `Noise4Sam_g0_t0.nidq.bin` or `Noise4Sam_g0_t0.imec0.lf.bin`
            # The filenames consist of 3 or 4 parts separated by `.`
            # 1. "Noise4Sam_g0_t0" will be the `name` variable. This choosen by the user
            #    at recording time.
            # 2. "_gt0_" will give the `seg_index` (here 0)
            # 3. "nidq" or "imec0" will give the `device` variable
            # 4. "lf" or "ap" will be the `signal_kind` variable
            # `stream_name` variable is the concatenation of `device.signal_kind`
            name = file.split('.')[0]
            r = re.findall(r'_g(\d*)_t', name)
            seg_index = int(r[0][0])
            device = file.split('.')[1]
            if 'imec' in device:
                signal_kind = file.split('.')[2]
                stream_name = device + '.' + signal_kind
                units = 'uV'
                # please note the 1e6 in gain for this uV

                # metad['imroTbl'] contain two gain per channel  AP and LF
                # except for the last fake channel
                per_channel_gain = np.ones(num_chan, dtype='float64')
                if 'imDatPrb_type' not in meta or meta['imDatPrb_type'] == '0':
                    # This work with NP 1.0 case with different metadata versions
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_3A.md#imec
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_3B1.md#imec
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_3B2.md#imec
                    if signal_kind == 'ap':
                        index_imroTbl = 3
                    elif signal_kind == 'lf':
                        index_imroTbl = 4
                    for c in range(num_chan - 1):
                        v = meta['imroTbl'][c].split(' ')[index_imroTbl]
                        per_channel_gain[c] = 1. / float(v)
                    gain_factor = float(meta['imAiRangeMax']) / 512
                    channel_gains = gain_factor * per_channel_gain * 1e6
                elif meta['imDatPrb_type'] in ('21', '24') and signal_kind == 'ap':
                    # This work with NP 2.0 case with different metadata versions
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_20.md#channel-entries-by-type
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_20.md#imec
                    # https://github.com/billkarsh/SpikeGLX/blob/gh-pages/Support/Metadata_30.md#imec
                    per_channel_gain[:-1] = 1 / 80.
                    gain_factor = float(meta['imAiRangeMax']) / 8192
                    channel_gains = gain_factor * per_channel_gain * 1e6
                else:
                    raise NotImplementedError('This meta file version of spikeglx'
                                              'is not implemented')
            else:
                signal_kind = ''
                stream_name = device
                units = 'V'
                channel_gains = np.ones(num_chan)

                # there are differents kinds of channels with different gain values
                mn, ma, xa, dw = [int(e) for e in meta['snsMnMaXaDw'].split(sep=',')]
                per_channel_gain = np.ones(num_chan, dtype='float64')
                per_channel_gain[0:mn] = 1. / float(meta['niMNGain'])
                per_channel_gain[mn:mn + ma] = 1. / float(meta['niMAGain'])
                # this scaling come from the code in this zip
                # https://billkarsh.github.io/SpikeGLX/Support/SpikeGLX_Datafile_Tools.zip
                # in file readSGLX.py line76
                # this is equivalent of 2**15
                gain_factor = float(meta['niAiRangeMax']) / 32768
                channel_gains = per_channel_gain * gain_factor

            info = {}
            info['name'] = name
            info['meta'] = meta
            info['meta_file'] = str(meta_filename)
            info['bin_file'] = str(bin_filename)
            for k in ('niSampRate', 'imSampRate'):
                if k in meta:
                    info['sampling_rate'] = float(meta[k])
            info['num_chan'] = num_chan

            info['sample_length'] = int(meta['fileSizeBytes']) // 2 // num_chan
            info['seg_index'] = seg_index
            info['device'] = device
            info['signal_kind'] = signal_kind
            info['stream_name'] = stream_name
            info['units'] = units
            info['channel_names'] = [txt.split(';')[0] for txt in meta['snsChanMap']]
            info['channel_gains'] = channel_gains
            info['channel_offsets'] = np.zeros(info['num_chan'])

            info_list.append(info)

    return info_list


def read_meta_file(meta_file):
    """parse the meta file"""
    with open(meta_file, mode='r') as f:
        lines = f.read().splitlines()

    info = {}
    # Fix taken from: https://github.com/SpikeInterface/probeinterface/blob/
    # 19d6518fbc67daca71aba5e99d8aa0d445b75eb7/probeinterface/io.py#L649-L662
    for line in lines:
        split_lines = line.split('=')
        if len(split_lines) != 2:
            continue
        k, v = split_lines
        if k.startswith('~'):
            # replace by the list
            k = k[1:]
            v = v[1:-1].split(')(')[1:]
        info[k] = v

    return info



import unittest

from neo.rawio.openephysbinaryrawio import OpenEphysBinaryRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestOpenEphysBinaryRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = OpenEphysBinaryRawIO
    entities_to_download = [
        'openephysbinary'
    ]
    entities_to_test = [
        'openephysbinary/v0.5.3_two_neuropixels_stream',
        'openephysbinary/v0.4.4.1_with_video_tracking',
        'openephysbinary/v0.5.x_two_nodes',
    ]


if __name__ == "__main__":
    unittest.main()



"""
This module implements the "new" binary OpenEphys format.
In this format channels are interleaved in one file.


See
https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html

Author: Julia Sprenger and Samuel Garcia
"""


import os
import re
import json

from pathlib import Path

import numpy as np

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
    _spike_channel_dtype, _event_channel_dtype)


class OpenEphysBinaryRawIO(BaseRawIO):
    """
    Handle several Blocks and several Segments.


    # Correspondencies
    Neo          OpenEphys
    block[n-1]   experiment[n]    New device start/stop
    segment[s-1] recording[s]     New recording start/stop

    This IO handles several signal streams.
    Special event (npy) data are represented as array_annotations.
    The current implementation does not handle spiking data, this will be added upon user request

    """
    extensions = []
    rawmode = 'one-dir'

    def __init__(self, dirname=''):
        BaseRawIO.__init__(self)
        self.dirname = dirname

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        all_streams, nb_block, nb_segment_per_block = explore_folder(self.dirname)

        sig_stream_names = sorted(list(all_streams[0][0]['continuous'].keys()))
        event_stream_names = sorted(list(all_streams[0][0]['events'].keys()))

        # first loop to reasign stream by "stream_index" instead of "stream_name"
        self._sig_streams = {}
        self._evt_streams = {}
        for block_index in range(nb_block):
            self._sig_streams[block_index] = {}
            self._evt_streams[block_index] = {}
            for seg_index in range(nb_segment_per_block[block_index]):
                self._sig_streams[block_index][seg_index] = {}
                self._evt_streams[block_index][seg_index] = {}
                for stream_index, stream_name in enumerate(sig_stream_names):
                    d = all_streams[block_index][seg_index]['continuous'][stream_name]
                    d['stream_name'] = stream_name
                    self._sig_streams[block_index][seg_index][stream_index] = d
                for i, stream_name in enumerate(event_stream_names):
                    d = all_streams[block_index][seg_index]['events'][stream_name]
                    d['stream_name'] = stream_name
                    self._evt_streams[block_index][seg_index][i] = d

        # signals zone
        # create signals channel map: several channel per stream
        signal_channels = []
        for stream_index, stream_name in enumerate(sig_stream_names):
            # stream_index is the index in vector sytream names
            stream_id = str(stream_index)
            d = self._sig_streams[0][0][stream_index]
            new_channels = []
            for chan_info in d['channels']:
                chan_id = chan_info['channel_name']
                new_channels.append((chan_info['channel_name'],
                    chan_id, float(d['sample_rate']), d['dtype'], chan_info['units'],
                    chan_info['bit_volts'], 0., stream_id))
            signal_channels.extend(new_channels)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        signal_streams = []
        for stream_index, stream_name in enumerate(sig_stream_names):
            stream_id = str(stream_index)
            signal_streams.append((stream_name, stream_id))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # create memmap for signals
        for block_index in range(nb_block):
            for seg_index in range(nb_segment_per_block[block_index]):
                for stream_index, d in self._sig_streams[block_index][seg_index].items():
                    num_channels = len(d['channels'])
                    memmap_sigs = np.memmap(d['raw_filename'], d['dtype'],
                                 order='C', mode='r').reshape(-1, num_channels)
                    d['memmap'] = memmap_sigs

        # events zone
        # channel map: one channel one stream
        event_channels = []
        for stream_ind, stream_name in enumerate(event_stream_names):
            d = self._evt_streams[0][0][stream_ind]
            event_channels.append((d['channel_name'], stream_ind, 'event'))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # create memmap
        for stream_ind, stream_name in enumerate(event_stream_names):
            # inject memmap loaded into main dict structure
            d = self._evt_streams[0][0][stream_ind]

            for name in _possible_event_stream_names:
                if name + '_npy' in d:
                    data = np.load(d[name + '_npy'], mmap_mode='r')
                    d[name] = data

            # check that events have timestamps
            assert 'timestamps' in d

            # for event the neo "label" will change depending the nature
            #  of event (ttl, text, binary)
            # and this is transform into unicode
            # all theses data are put in event array annotations
            if 'text' in d:
                # text case
                d['labels'] = d['text'].astype('U')
            elif 'metadata' in d:
                # binary case
                d['labels'] = d['channels'].astype('U')
            elif 'channels' in d:
                # ttl case use channels
                d['labels'] = d['channels'].astype('U')
            else:
                raise ValueError(f'There is no possible labels for this event: {stream_name}')

        # no spike read yet
        # can be implemented on user demand
        spike_channels = np.array([], dtype=_spike_channel_dtype)

        # loop for t_start/t_stop on segment browse all object
        self._t_start_segments = {}
        self._t_stop_segments = {}
        for block_index in range(nb_block):
            self._t_start_segments[block_index] = {}
            self._t_stop_segments[block_index] = {}
            for seg_index in range(nb_segment_per_block[block_index]):
                global_t_start = None
                global_t_stop = None

                # loop over signals
                for stream_index, d in self._sig_streams[block_index][seg_index].items():
                    t_start = d['t_start']
                    dur = d['memmap'].shape[0] / float(d['sample_rate'])
                    t_stop = t_start + dur
                    if global_t_start is None or global_t_start > t_start:
                        global_t_start = t_start
                    if global_t_stop is None or global_t_stop < t_stop:
                        global_t_stop = t_stop

                # loop over events
                for stream_index, stream_name in enumerate(event_stream_names):
                    d = self._evt_streams[0][0][stream_index]
                    if d['timestamps'].size == 0:
                        continue
                    t_start = d['timestamps'][0] / d['sample_rate']
                    t_stop = d['timestamps'][-1] / d['sample_rate']
                    if global_t_start is None or global_t_start > t_start:
                        global_t_start = t_start
                    if global_t_stop is None or global_t_stop < t_stop:
                        global_t_stop = t_stop

                self._t_start_segments[block_index][seg_index] = global_t_start
                self._t_stop_segments[block_index][seg_index] = global_t_stop

        # main header
        self.header = {}
        self.header['nb_block'] = nb_block
        self.header['nb_segment'] = nb_segment_per_block
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        # Annotate some objects from continuous files
        self._generate_minimal_annotations()
        for block_index in range(nb_block):
            bl_ann = self.raw_annotations['blocks'][block_index]
            for seg_index in range(nb_segment_per_block[block_index]):
                seg_ann = bl_ann['segments'][seg_index]

                # array annotations for signal channels
                for stream_index, stream_name in enumerate(sig_stream_names):
                    sig_ann = seg_ann['signals'][stream_index]
                    d = self._sig_streams[0][0][stream_index]
                    for k in ('identifier', 'history', 'source_processor_index',
                              'recorded_processor_index'):
                        if k in d['channels'][0]:
                            values = np.array([chan_info[k] for chan_info in d['channels']])
                            sig_ann['__array_annotations__'][k] = values

                # array annotations for event channels
                # use other possible data in _possible_event_stream_names
                for stream_index, stream_name in enumerate(event_stream_names):
                    ev_ann = seg_ann['events'][stream_index]
                    d = self._evt_streams[0][0][stream_index]
                    for k in _possible_event_stream_names:
                        if k in ('timestamps', ):
                            continue
                        if k in d:
                            # split custom dtypes into separate annotations
                            if d[k].dtype.names:
                                for name in d[k].dtype.names:
                                    ev_ann['__array_annotations__'][name] = d[k][name].flatten()
                            else:
                                ev_ann['__array_annotations__'][k] = d[k]

    def _segment_t_start(self, block_index, seg_index):
        return self._t_start_segments[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stop_segments[block_index][seg_index]

    def _channels_to_group_id(self, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        channels = self.header['signal_channels']
        group_ids = channels[channel_indexes]['group_id']
        assert np.unique(group_ids).size == 1
        group_id = group_ids[0]
        return group_id

    def _get_signal_size(self, block_index, seg_index, stream_index):
        sigs = self._sig_streams[block_index][seg_index][stream_index]['memmap']
        return sigs.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        t_start = self._sig_streams[block_index][seg_index][stream_index]['t_start']
        return t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        sigs = self._sig_streams[block_index][seg_index][stream_index]['memmap']
        sigs = sigs[i_start:i_stop, :]
        if channel_indexes is not None:
            sigs = sigs[:, channel_indexes]
        return sigs

    def _spike_count(self, block_index, seg_index, unit_index):
        pass

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        pass

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        pass

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        pass

    def _event_count(self, block_index, seg_index, event_channel_index):
        d = self._evt_streams[0][0][event_channel_index]
        return d['timestamps'].size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        d = self._evt_streams[0][0][event_channel_index]
        timestamps = d['timestamps']
        durations = None
        labels = d['labels']

        # slice it if needed
        if t_start is not None:
            ind_start = int(t_start * d['sample_rate'])
            mask = timestamps >= ind_start
            timestamps = timestamps[mask]
            labels = labels[mask]
        if t_stop is not None:
            ind_stop = int(t_stop * d['sample_rate'])
            mask = timestamps < ind_stop
            timestamps = timestamps[mask]
            labels = labels[mask]
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        d = self._evt_streams[0][0][event_channel_index]
        event_times = event_timestamps.astype(dtype) / float(d['sample_rate'])
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype):
        pass


_possible_event_stream_names = ('timestamps', 'channels', 'text',
        'full_word', 'channel_states', 'data_array', 'metadata')


def explore_folder(dirname):
    """
    Exploring the OpenEphys folder structure and structure.oebin

    Returns nested dictionary structure:
    [block_index][seg_index][stream_type][stream_information]
    where
    - node_name is the open ephys node id
    - block_index is the neo Block index
    - segment_index is the neo Segment index
    - stream_type can be 'continuous'/'events'/'spikes'
    - stream_information is a dictionionary containing e.g. the sampling rate

    Parmeters
    ---------
    dirname (str): Root folder of the dataset

    Returns
    -------
    nested dictionaries containing structure and stream information
    """
    nb_block = 0
    nb_segment_per_block = []
    # nested dictionary: block_index > seg_index > data_type > stream_name
    all_streams = {}
    for root, dirs, files in os.walk(dirname):
        for file in files:
            if not file == 'structure.oebin':
                continue
            root = Path(root)

            node_name = root.parents[1].stem
            if not node_name.startswith('Record'):
                # before version 5.x.x there was not multi Node recording
                # so no node_name
                node_name = ''

            block_index = int(root.parents[0].stem.lower().replace('experiment', '')) - 1
            if block_index not in all_streams:
                all_streams[block_index] = {}
                if block_index >= nb_block:
                    nb_block = block_index + 1
                    nb_segment_per_block.append(0)

            seg_index = int(root.stem.replace('recording', '')) - 1
            if seg_index not in all_streams[block_index]:
                all_streams[block_index][seg_index] = {
                    'continuous': {},
                    'events': {},
                    'spikes': {},
                }
                if seg_index >= nb_segment_per_block[block_index]:
                    nb_segment_per_block[block_index] = seg_index + 1

            # metadata
            with open(root / 'structure.oebin', encoding='utf8', mode='r') as f:
                structure = json.load(f)

            if (root / 'continuous').exists() and len(structure['continuous']) > 0:
                for d in structure['continuous']:
                    # when multi Record Node the stream name also contains
                    # the node name to make it unique
                    stream_name = node_name + '#' + d['folder_name']

                    raw_filename = root / 'continuous' / d['folder_name'] / 'continuous.dat'

                    timestamp_file = root / 'continuous' / d['folder_name'] / 'timestamps.npy'
                    timestamps = np.load(str(timestamp_file), mmap_mode='r')
                    timestamp0 = timestamps[0]
                    t_start = timestamp0 / d['sample_rate']

                    # TODO for later : gap checking
                    signal_stream = d.copy()
                    signal_stream['raw_filename'] = str(raw_filename)
                    signal_stream['dtype'] = 'int16'
                    signal_stream['timestamp0'] = timestamp0
                    signal_stream['t_start'] = t_start

                    all_streams[block_index][seg_index]['continuous'][stream_name] = signal_stream

            if (root / 'events').exists() and len(structure['events']) > 0:
                for d in structure['events']:
                    stream_name = node_name + '#' + d['folder_name']

                    event_stream = d.copy()
                    for name in _possible_event_stream_names:
                        npz_filename = root / 'events' / d['folder_name'] / f'{name}.npy'
                        if npz_filename.is_file():
                            event_stream[f'{name}_npy'] = str(npz_filename)

                    all_streams[block_index][seg_index]['events'][stream_name] = event_stream

    # TODO for later: check stream / channel consistency across segment

    return all_streams, nb_block, nb_segment_per_block



"""
Tests of neo.rawio.spikeglxrawio
"""

import unittest

from neo.rawio.spikeglxrawio import SpikeGLXRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestSpikeGLXRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = SpikeGLXRawIO
    entities_to_download = [
        'spikeglx'
    ]
    entities_to_test = [
        'spikeglx/Noise4Sam_g0',
        'spikeglx/TEST_20210920_0_g0'
    ]

    def test_with_location(self):
        rawio = SpikeGLXRawIO(self.get_local_path('spikeglx/Noise4Sam_g0'), load_channel_location=True)
        rawio.parse_header()
        # one of the stream have channel location
        have_location = []
        for sig_anotations in rawio.raw_annotations['blocks'][0]['segments'][0]['signals']:
            have_location.append('channel_location_0' in sig_anotations['__array_annotations__'])
        assert any(have_location)


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.mearecrawio import MEArecRawIO


class MEArecIO(MEArecRawIO, BaseFromRaw):
    __doc__ = MEArecRawIO.__doc__
    mode = 'file'

    def __init__(self, filename):
        MEArecRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Tests of neo.rawio.BiocamRawIO
"""

import unittest

from neo.rawio.biocamrawio import BiocamRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestBiocamRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = BiocamRawIO

    entities_to_download = [
        'biocam/biocam_hw3.0_fw1.6.brw'
        ]

    entities_to_download = [
        'biocam',
    ]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.spikeglxrawio import SpikeGLXRawIO


class SpikeGLXIO(SpikeGLXRawIO, BaseFromRaw):
    __doc__ = SpikeGLXRawIO.__doc__
    mode = 'dir'

    def __init__(self, dirname, load_sync_channel=False, load_channel_location=False):
        SpikeGLXRawIO.__init__(self, dirname=dirname,
            load_sync_channel=load_sync_channel,
            load_channel_location=load_channel_location)
        BaseFromRaw.__init__(self, dirname)



"""
Common tools that are useful for neo.io object tests
"""

import logging
import os

from neo.utils import HAVE_DATALAD

logger = logging.getLogger("neo.test")


def can_use_network():
    """
    Return True if network access is allowed
    """
    if not HAVE_DATALAD:
        return False
    if os.environ.get('NOSETESTS_NO_NETWORK', False):
        return False
    if os.environ.get('TRAVIS') == 'true':
        return False
    return True



"""
Tests of neo.io.asciispiketrainio
"""

import unittest

from neo.io import AsciiSpikeTrainIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestAsciiSpikeTrainIO(BaseTestIO, unittest.TestCase, ):
    ioclass = AsciiSpikeTrainIO
    entities_to_download = [
        'asciispiketrain'
    ]
    entities_to_test = [
        'asciispiketrain/File_ascii_spiketrain_1.txt',
    ]


if __name__ == "__main__":
    unittest.main()



"""
Class for reading data from a MEArec simulated data.

See:
https://mearec.readthedocs.io/en/latest/
https://github.com/alejoe91/MEArec
https://link.springer.com/article/10.1007/s12021-020-09467-7

Author : Alessio Buccino
"""

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                _spike_channel_dtype, _event_channel_dtype)

import numpy as np
from copy import deepcopy


class MEArecRawIO(BaseRawIO):
    """
    Class for "reading" fake data from a MEArec file.

    Usage:
        >>> import neo.rawio
        >>> r = neo.rawio.MEArecRawIO(filename='mearec.h5')
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0, seg_index=0,
                            i_start=0, i_stop=1024,  channel_names=channel_names)
        >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk, dtype='float64',
                            channel_indexes=[0, 3, 6])
        >>> spike_timestamp = reader.spike_timestamps(unit_index=0, t_start=None, t_stop=None)
        >>> spike_times = reader.rescale_spike_timestamp(spike_timestamp, 'float64')

    """
    extensions = ['h5']
    rawmode = 'one-file'

    def __init__(self, filename=''):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        try:
            import MEArec as mr
            HAVE_MEAREC = True
        except ImportError:
            HAVE_MEAREC = False
        assert HAVE_MEAREC, 'MEArec is not installed'
        self._recgen = mr.load_recordings(recordings=self.filename, return_h5_objects=True,
                                          check_suffix=False,
                                          load=['recordings', 'spiketrains', 'channel_positions'],
                                          load_waveforms=False)
        self._sampling_rate = self._recgen.info['recordings']['fs']
        self._recordings = self._recgen.recordings
        self._num_frames, self._num_channels = self._recordings.shape

        signal_streams = np.array([('Signals', '0')], dtype=_signal_stream_dtype)

        sig_channels = []
        for c in range(self._num_channels):
            ch_name = 'ch{}'.format(c)
            chan_id = str(c + 1)
            sr = self._sampling_rate  # Hz
            dtype = self._recordings.dtype
            units = 'uV'
            gain = 1.
            offset = 0.
            stream_id = '0'
            sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # creating units channels
        spike_channels = []
        self._spiketrains = self._recgen.spiketrains
        for c in range(len(self._spiketrains)):
            unit_name = 'unit{}'.format(c)
            unit_id = '#{}'.format(c)
            # if spiketrains[c].waveforms is not None:
            wf_units = ''
            wf_gain = 1.
            wf_offset = 0.
            wf_left_sweep = 0
            wf_sampling_rate = self._sampling_rate
            spike_channels.append((unit_name, unit_id, wf_units, wf_gain,
                                  wf_offset, wf_left_sweep, wf_sampling_rate))
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [1]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = sig_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        self._generate_minimal_annotations()
        for block_index in range(1):
            bl_ann = self.raw_annotations['blocks'][block_index]
            bl_ann['mearec_info'] = deepcopy(self._recgen.info)

    def _segment_t_start(self, block_index, seg_index):
        all_starts = [[0.]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._num_frames / self._sampling_rate
        all_stops = [[t_stop]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        assert stream_index == 0
        return self._num_frames

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        assert stream_index == 0
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._num_frames

        if channel_indexes is None:
            channel_indexes = slice(self._num_channels)
        if isinstance(channel_indexes, slice):
            raw_signals = self._recgen.recordings[i_start:i_stop, channel_indexes]
        else:
            # sort channels because h5py neeeds sorted indexes
            if np.any(np.diff(channel_indexes) < 0):
                sorted_channel_indexes = np.sort(channel_indexes)
                sorted_idx = np.array([list(sorted_channel_indexes).index(ch)
                                       for ch in channel_indexes])
                raw_signals = self._recgen.recordings[i_start:i_stop, sorted_channel_indexes]
                raw_signals = raw_signals[:, sorted_idx]
            else:
                raw_signals = self._recgen.recordings[i_start:i_stop, channel_indexes]
        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):
        return len(self._spiketrains[unit_index])

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        spike_timestamps = self._spiketrains[unit_index].times.magnitude
        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)
        timestamp_idxs = np.where((spike_timestamps >= t_start) & (spike_timestamps < t_stop))

        return spike_timestamps[timestamp_idxs]

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        return spike_timestamps.astype(dtype)

    def _get_spike_raw_waveforms(self, block_index, seg_index,
                                 spike_channel_index, t_start, t_stop):
        return None



import unittest

from neo.rawio.cedrawio import CedRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestCedRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = CedRawIO
    entities_to_test = [
        'spike2/m365_1sec.smrx',
        'spike2/File_spike2_1.smr',
        'spike2/Two-mice-bigfile-test000.smr'
    ]
    entities_to_download = [
        'spike2'
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.elanrawio import ElanRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestElanRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = ElanRawIO
    entities_to_test = [
        'elan/File_elan_1.eeg'
    ]
    entities_to_download = [
        'elan',
    ]


if __name__ == "__main__":
    unittest.main()



'''
Utils package
'''

from .misc import (get_events, get_epochs, add_epoch,
    match_events, cut_block_by_epochs, cut_segment_by_epoch,
    is_block_rawio_compatible)
from .datasets import (download_dataset,
    get_local_testing_data_folder, HAVE_DATALAD)



import unittest

import numpy as np

from neo.rawio.neuralynxrawio.neuralynxrawio import NeuralynxRawIO
from neo.rawio.neuralynxrawio.nlxheader import NlxHeader
from neo.rawio.neuralynxrawio.ncssections import (NcsSection, NcsSections,
                                                  NcsSectionsFactory)
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

import logging

logging.getLogger().setLevel(logging.INFO)


class TestNeuralynxRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = NeuralynxRawIO
    entities_to_download = [
        'neuralynx'
    ]
    entities_to_test = [
        'neuralynx/BML/original_data',
        'neuralynx/BML_unfilledsplit/original_data',
        'neuralynx/Cheetah_v1.1.0/original_data',
        'neuralynx/Cheetah_v4.0.2/original_data',
        'neuralynx/Cheetah_v5.4.0/original_data',
        'neuralynx/Cheetah_v5.5.1/original_data',
        'neuralynx/Cheetah_v5.6.3/original_data',
        'neuralynx/Cheetah_v5.7.4/original_data',
        'neuralynx/Cheetah_v6.3.2/incomplete_blocks']

    def test_scan_ncs_files(self):

        # Test BML style of Ncs files, similar to PRE4 but with fractional frequency
        # in the header and fractional microsPerSamp, which is then rounded as appropriate
        # in each record.
        rawio = NeuralynxRawIO(self.get_local_path('neuralynx/BML/original_data'))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 1)
        self.assertListEqual(rawio._timestamp_limits, [(0, 192000)])
        self.assertEqual(rawio._sigs_length[0], 4608)
        self.assertEqual(rawio._sigs_t_start[0], 0)
        self.assertEqual(rawio._sigs_t_stop[0], 0.192)
        self.assertEqual(len(rawio._sigs_memmaps), 1)

        # Test Cheetah 4.0.2, which is PRE4 type with frequency in header and
        # no microsPerSamp. Number of microseconds per sample in file is inverse of
        # sampling frequency in header trucated to microseconds.
        rawio = NeuralynxRawIO(self.get_local_path('neuralynx/Cheetah_v4.0.2/original_data'))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 1)
        self.assertListEqual(rawio._timestamp_limits, [(266982936, 267162136)])
        self.assertEqual(rawio._sigs_length[0], 5120)
        self.assertEqual(rawio._sigs_t_start[0], 266.982936)
        self.assertEqual(rawio._sigs_t_stop[0], 267.162136)
        self.assertEqual(len(rawio._sigs_memmaps), 1)

        # Test Cheetah 5.5.1, which is DigitalLynxSX and has two blocks of records
        # with a fairly large gap.
        rawio = NeuralynxRawIO(self.get_local_path('neuralynx/Cheetah_v5.5.1/original_data'))
        rawio.parse_header()
        # test values here from direct inspection of .ncs files
        self.assertEqual(rawio._nb_segment, 2)
        self.assertListEqual(rawio._timestamp_limits, [(26122557633, 26162525633),
                                                       (26366360633, 26379704633)])
        self.assertListEqual(rawio._sigs_length, [1278976, 427008])
        self.assertListEqual(rawio._sigs_t_stop, [26162.525633, 26379.704633])
        self.assertListEqual(rawio._sigs_t_start, [26122.557633, 26366.360633])
        self.assertEqual(len(rawio._sigs_memmaps), 2)  # check only that there are 2 memmaps

        # Test Cheetah 6.3.2, the incomplete_blocks test. This is a DigitalLynxSX with
        # three blocks of records. Gaps are on the order of 60 microseconds or so.
        rawio = NeuralynxRawIO(self.get_local_path('neuralynx/Cheetah_v6.3.2/incomplete_blocks'))
        rawio.parse_header()
        # test values here from direct inspection of .ncs file, except for 3rd block
        # t_stop, which is extended due to events past the last block of ncs records.
        self.assertEqual(rawio._nb_segment, 3)
        self.assertListEqual(rawio._timestamp_limits, [(8408806811, 8427831990),
                                                       (8427832053, 8487768498),
                                                       (8487768561, 8515816549)])
        self.assertListEqual(rawio._sigs_length, [608806, 1917967, 897536])
        self.assertListEqual(rawio._sigs_t_stop, [8427.831990, 8487.768498, 8515.816549])
        self.assertListEqual(rawio._sigs_t_start, [8408.806811, 8427.832053, 8487.768561])
        self.assertEqual(len(rawio._sigs_memmaps), 3)  # check only that there are 3 memmaps

    def test_single_file_mode(self):
        """
        Tests reading of single files.
        """

        # test single analog signal channel
        fname = self.get_local_path('neuralynx/Cheetah_v5.6.3/original_data/CSC1.ncs')
        rawio = NeuralynxRawIO(filename=fname)
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 0)
        sigHdrs = rawio.header['signal_channels']
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], 'CSC1')
        self.assertEqual(sigHdrs[0][1], '58')
        self.assertEqual(len(rawio.header['spike_channels']), 0)
        self.assertEqual(len(rawio.header['event_channels']), 0)

        # test one single electrode channel
        fname = self.get_local_path('neuralynx/Cheetah_v5.5.1/original_data/STet3a.nse')
        rawio = NeuralynxRawIO(filename=fname)
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 1)
        self.assertEqual(len(rawio.ncs_filenames), 0)
        self.assertEqual(len(rawio.nev_filenames), 0)
        seHdrs = rawio.header['spike_channels']
        self.assertEqual(len(seHdrs), 1)
        self.assertEqual(seHdrs[0][0], 'chSTet3a#8#0')
        self.assertEqual(seHdrs[0][1], '0')
        self.assertEqual(len(rawio.header['signal_channels']), 0)
        self.assertEqual(len(rawio.header['event_channels']), 0)

    def test_exclude_filenames(self):
        # exclude single ncs file from session
        dname = self.get_local_path('neuralynx/Cheetah_v5.6.3/original_data/')
        rawio = NeuralynxRawIO(dirname=dname, exclude_filename='CSC2.ncs')
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 1)
        sigHdrs = rawio.header['signal_channels']
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], 'CSC1')
        self.assertEqual(sigHdrs[0][1], '58')
        self.assertEqual(len(rawio.header['spike_channels']), 8)
        self.assertEqual(len(rawio.header['event_channels']), 2)

        # exclude multiple files from session
        rawio = NeuralynxRawIO(dirname=dname, exclude_filename=['Events.nev', 'CSC2.ncs'])
        rawio.parse_header()

        self.assertEqual(rawio._nb_segment, 2)
        self.assertEqual(len(rawio.ncs_filenames), 1)
        self.assertEqual(len(rawio.nev_filenames), 0)
        sigHdrs = rawio.header['signal_channels']
        self.assertEqual(sigHdrs.size, 1)
        self.assertEqual(sigHdrs[0][0], 'CSC1')
        self.assertEqual(sigHdrs[0][1], '58')
        self.assertEqual(len(rawio.header['spike_channels']), 8)
        self.assertEqual(len(rawio.header['event_channels']), 0)

class TestNcsRecordingType(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test of decoding of NlxHeader for type of recording.
    """
    entities_to_test = []

    ncsTypeTestFiles = [
        ('neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs', 'PRE4'),
        ('neuralynx/Cheetah_v5.4.0/original_data/CSC5_trunc.Ncs', 'DIGITALLYNX'),
        ('neuralynx/Cheetah_v5.5.1/original_data/STet3a.nse', 'DIGITALLYNXSX'),
        ('neuralynx/Cheetah_v5.5.1/original_data/Tet3a.ncs', 'DIGITALLYNXSX'),
        ('neuralynx/Cheetah_v5.6.3/original_data/CSC1.ncs', 'DIGITALLYNXSX'),
        ('neuralynx/Cheetah_v5.6.3/original_data/TT1.ntt', 'DIGITALLYNXSX'),
        ('neuralynx/Cheetah_v5.7.4/original_data/CSC1.ncs', 'DIGITALLYNXSX'),
        ('neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs', 'DIGITALLYNXSX')]

    def test_recording_types(self):

        for typeTest in self.ncsTypeTestFiles:

            filename = self.get_local_path(typeTest[0])
            hdr = NlxHeader(filename)
            self.assertEqual(hdr.type_of_recording(), typeTest[1])


class TestNcsSectionsFactory(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test building NcsBlocks for files of different revisions.
    """
    entities_to_test = []

    def test_ncsblocks_partial(self):
        filename = self.get_local_path('neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        self.assertEqual(data0.shape[0], 6690)
        self.assertEqual(data0['timestamp'][6689], 8515800549)  # timestamp of last record

        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(nb.sampFreqUsed, 32000.012813673042)
        self.assertEqual(nb.microsPerSampUsed, 31.249987486652431)
        self.assertListEqual([blk.startRec for blk in nb.sects], [0, 1190, 4937])
        self.assertListEqual([blk.endRec for blk in nb.sects], [1189, 4936, 6689])

    def test_build_given_actual_frequency(self):

        # Test early files where the frequency listed in the header is
        # floor(1e6/(actual number of microseconds between samples)
        filename = self.get_local_path('neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        ncsBlocks = NcsSections()
        ncsBlocks.sampFreqUsed = 1 / (35e-6)
        ncsBlocks.microsPerSampUsed = 35
        ncsBlocks = NcsSectionsFactory._buildGivenActualFrequency(data0, ncsBlocks.sampFreqUsed,
                                                                  27789)
        self.assertEqual(len(ncsBlocks.sects), 1)
        self.assertEqual(ncsBlocks.sects[0].startRec, 0)
        self.assertEqual(ncsBlocks.sects[0].endRec, 9)

    def test_build_using_header_and_scanning(self):

        # Test early files where the frequency listed in the header is
        # floor(1e6/(actual number of microseconds between samples)
        filename = self.get_local_path('neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs')
        hdr = NlxHeader(filename)
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)

        self.assertEqual(nb.sampFreqUsed, 1 / 35e-6)
        self.assertEqual(nb.microsPerSampUsed, 35)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startRec, 0)
        self.assertEqual(nb.sects[0].endRec, 9)

        # test Cheetah 5.5.1, which is DigitalLynxSX and has two blocks of records
        # with a fairly large gap
        filename = self.get_local_path('neuralynx/Cheetah_v5.5.1/original_data/Tet3a.ncs')
        hdr = NlxHeader(filename)
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(nb.sampFreqUsed, 32000)
        self.assertEqual(nb.microsPerSampUsed, 31.25)
        self.assertEqual(len(nb.sects), 2)
        self.assertListEqual([blk.startRec for blk in nb.sects], [0, 2498])
        self.assertListEqual([blk.endRec for blk in nb.sects], [2497, 3331])

    def test_block_start_and_end_times(self):
        # digitallynxsx version to exercise the _parseForMaxGap function with multiple blocks
        filename = self.get_local_path('neuralynx/Cheetah_v6.3.2/incomplete_blocks/CSC1_reduced.ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertListEqual([blk.startTime for blk in nb.sects], [8408806811, 8427832053,
                                                                   8487768561])
        self.assertListEqual([blk.endTime for blk in nb.sects], [8427831990, 8487768498,
                                                                 8515816549])

        # digitallynxsx with single block of records to exercise path in _buildForMaxGap
        filename = self.get_local_path('neuralynx/Cheetah_v1.1.0/original_data/CSC67_trunc.Ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startTime, 253293161778)
        self.assertEqual(nb.sects[0].endTime, 253293349278)

        # PRE4 version with single block of records to exercise path in _buildGivenActualFrequency
        filename = self.get_local_path('neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 1)
        self.assertEqual(nb.sects[0].startTime, 266982936)
        self.assertEqual(nb.sects[0].endTime, 267162136)

        # BML style with two blocks of records and one partially filled record to exercise
        # _parseGivenActualFrequency
        filename = self.get_local_path(
            'neuralynx/BML_unfilledsplit/original_data/unfilledSplitRecords.Ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr = NlxHeader(filename)
        nb = NcsSectionsFactory.build_for_ncs_file(data0, hdr)
        self.assertEqual(len(nb.sects), 2)
        self.assertListEqual([blk.startTime for blk in nb.sects], [1837623129, 6132625241])
        self.assertListEqual([blk.endTime for blk in nb.sects], [1837651009, 6132642649])

    def test_block_verify(self):
        # check that file verifies against itself for single block
        filename = self.get_local_path('neuralynx/Cheetah_v4.0.2/original_data/CSC14_trunc.Ncs')
        data0 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr0 = NlxHeader(filename)
        nb0 = NcsSectionsFactory.build_for_ncs_file(data0, hdr0)

        self.assertTrue(NcsSectionsFactory._verifySectionsStructure(data0, nb0))

        # check that fails against file with two blocks
        filename = self.get_local_path(
            'neuralynx/BML_unfilledsplit/original_data/unfilledSplitRecords.Ncs')
        data1 = np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode='r',
                          offset=NlxHeader.HEADER_SIZE)
        hdr1 = NlxHeader(filename)
        nb1 = NcsSectionsFactory.build_for_ncs_file(data1, hdr1)

        self.assertFalse(NcsSectionsFactory._verifySectionsStructure(data1, nb0))

        # check that two blocks verify against self
        self.assertTrue(NcsSectionsFactory._verifySectionsStructure(data1, nb1))


class TestNcsSections(TestNeuralynxRawIO, unittest.TestCase):
    """
    Test building NcsBlocks for files of different revisions.
    """
    entities_to_test = []

    def test_equality(self):
        ns0 = NcsSections()
        ns1 = NcsSections()

        ns0.microsPerSampUsed = 1
        ns1.microsPerSampUsed = 1
        ns0.sampFreqUsed = 300
        ns1.sampFreqUsed = 300

        self.assertEqual(ns0, ns1)

        # add sections
        ns0.sects = [NcsSection(0, 0, 100, 100, 10)]
        ns1.sects = [NcsSection(0, 0, 100, 100, 10)]

        self.assertEqual(ns0, ns1)

        # check inequality for different attributes
        # different number of sections
        ns0.sects.append(NcsSection(0, 0, 100, 100, 10))
        self.assertNotEqual(ns0, ns1)

        # different section attributes
        ns0.sects = [NcsSection(0, 0, 200, 200, 10)]
        self.assertNotEqual(ns0, ns1)

        # different attributes
        ns0.sampFreqUsed = 400
        self.assertNotEqual(ns0, ns1)


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.neuroscopeio
"""

import unittest

from neo.io import NeuroScopeIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNeuroScopeIO(BaseTestIO, unittest.TestCase, ):
    ioclass = NeuroScopeIO
    entities_to_download = [
        'neuroscope'
    ]
    entities_to_test = ['neuroscope/test1/test1.xml']


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.kwikio
"""

import unittest

try:
    import h5py

    HAVE_H5PY = True
except ImportError:
    HAVE_H5PY = False
from neo.io import kwikio
from neo.test.iotest.common_io_test import BaseTestIO


@unittest.skipUnless(HAVE_H5PY, "requires h5py")
@unittest.skipUnless(kwikio.HAVE_KWIK, "requires klusta")
class TestKwikIO(BaseTestIO, unittest.TestCase):
    ioclass = kwikio.KwikIO
    entities_to_download = [
        'kwik'
    ]
    entities_to_test = [
        'kwik/neo.kwik'
    ]


if __name__ == "__main__":
    unittest.main()



"""

"""

import unittest

from neo.rawio.brainvisionrawio import BrainVisionRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestBrainVisionRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = BrainVisionRawIO
    entities_to_test = [
        'brainvision/File_brainvision_1.vhdr',
        'brainvision/File_brainvision_2.vhdr',
        'brainvision/File_brainvision_3_float32.vhdr',
        'brainvision/File_brainvision_3_int16.vhdr',
        'brainvision/File_brainvision_3_int32.vhdr',
    ]

    entities_to_download = [
        'brainvision'
    ]


if __name__ == "__main__":
    unittest.main()
